import mimetypes as _mimetypes
import os
from mimetypes import *

from .. import WSB_USER_DIR

__all__ = _mimetypes.__all__


# add custom user MIME types mapping
_mimetypes.knownfiles += [os.path.join(WSB_USER_DIR, 'mime.types')]

# WebScrapBook related
_mimetypes.add_type('application/html+zip', '.htz')
_mimetypes.add_type('application/x-maff', '.maff')
_mimetypes.add_type('application/wsba+zip', '.wsba')

# Some common types
_mimetypes.add_type('text/markdown', '.md')
_mimetypes.add_type('text/markdown', '.mkd')
_mimetypes.add_type('text/markdown', '.mkdn')
_mimetypes.add_type('text/markdown', '.mdwn')
_mimetypes.add_type('text/markdown', '.mdown')
_mimetypes.add_type('text/markdown', '.markdown')
_mimetypes.add_type('application/rss+xml', '.rss')
_mimetypes.add_type('application/atom+xml', '.atom')
_mimetypes.add_type('font/woff', '.woff')
_mimetypes.add_type('font/woff2', '.woff2')
_mimetypes.add_type('image/webp', '.webp')
_mimetypes.add_type('audio/weba', '.weba')
_mimetypes.add_type('video/webm', '.webm')
_mimetypes.add_type('audio/ogg', '.oga')
_mimetypes.add_type('video/ogg', '.ogv')
_mimetypes.add_type('application/ogg', '.ogx')  # IANA
_mimetypes.add_type('application/ogg', '.ogg')  # MAFF
_mimetypes.add_type('text/vtt', '.vtt')
_mimetypes.add_type('application/java-archive', '.jar')
_mimetypes.add_type('application/java-vm', '.class')
_mimetypes.add_type('application/epub+zip', '.epub')
_mimetypes.add_type('application/x-7z-compressed', '.7z')
_mimetypes.add_type('application/vnd.rar', '.rar')

# .js is mapped to application/x-javascript in some mime types sources
_mimetypes.add_type('application/javascript', '.js')

# .bmp is mapped to image/x-ms-bmp on POSIX
# ref: https://bugs.python.org/issue42028
_mimetypes.add_type('image/bmp', '.bmp')

# .ico is mapped to image/vnd.microsoft.icon on POSIX
_mimetypes.add_type('image/x-icon', '.ico')

# .zip is mapped to application/x-zip-compressed on Windows
_mimetypes.add_type('application/zip', '.zip')



"""HTML Parser
"""
import codecs
import html
import html.parser
import re
from urllib.parse import urljoin
from urllib.request import pathname2url

from . import util

REGEX_ASCII_WHITESPACES = re.compile(r'[ \t\n\r\f]+')

VOID_ELEMENTS = {'area', 'base', 'br', 'col', 'embed', 'hr', 'img', 'input', 'link', 'meta', 'param', 'source', 'track', 'wbr', 'frame'}
FOREIGN_ELEMENTS = {'svg', 'math'}


class Error(Exception):
    pass


class Markup:
    """Class representing an HTML markup.
    """
    SHOWN_FLAG_ATTRS = ['ignored', 'hidden']
    SHOWN_ATTRS = {
        'pi': ['data'],
        'decl': ['data'],
        'comment': ['data'],
        'starttag': ['tag', 'attrs', 'is_self_end'],
        'endtag': ['tag', 'attrs'],
        'data': ['data', 'is_cdata', 'convert_charrefs'],
        'entityref': ['name'],
        'charref': ['name'],
    }

    def __init__(self, **kwargs):
        # general
        self.is_xhtml = False
        self.type = None
        self.src = None  # corresponding source text
        self.ignored = False  # should be treated as non-exist in the DOM
        self.hidden = False  # should not be output

        # starttag, endtag, startendtag
        self.tag = None
        self.attrs = None
        self.starttag = None
        self.endtag = None
        self.is_self_end = False

        # entityref, charref
        self.name = None

        # pi, decl, comment, data
        self.data = None
        self.is_unknown = False

        # data
        self.convert_charrefs = False
        self.is_cdata = False

        # merge passed kwargs
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def __repr__(self):
        attrs = []
        for k in self.SHOWN_ATTRS[self.type]:
            attrs.append(f'{k}={repr(getattr(self, k))}')

        # try:
            # attrs.append(f'id={id(self)}')
        # except AttributeError:
            # pass

        # try:
            # attrs.append(f'starttag={id(self.starttag)}')
        # except AttributeError:
            # pass

        # try:
            # attrs.append(f'endtag={id(self.endtag)}')
        # except AttributeError:
            # pass

        for k in self.SHOWN_FLAG_ATTRS:
            if getattr(self, k):
                attrs.append(k)

        return f'{self.type}({", ".join(attrs)})'

    def __str__(self):
        # use recorded src if exists
        if self.src is not None:
            return self.src

        if self.type == 'data':
            if self.is_cdata:
                return self.data
            if self.convert_charrefs:
                return html.escape(self.data, quote=False)
            return self.data
        elif self.type == 'starttag':
            if self.is_xhtml:
                attrs = ' '.join((f'{k}="{html.escape(k if v is None else v)}"') for k, v in self.attrs)
            else:
                attrs = ' '.join((k if v is None else f'{k}="{html.escape(v)}"') for k, v in self.attrs)
            attrs = ' ' + self._rewrite_attrs(attrs) if attrs else ''
            return f'<{self.tag}{attrs}{" /" if self.is_self_end else ""}>'
        elif self.type == 'endtag':
            return f'</{self.tag}>'
        elif self.type == 'pi':
            return f'<?{self.data}>'
        elif self.type == 'decl':
            return f'<!{self.data}>'
        elif self.type == 'comment':
            return f'<!--{self.data}-->'
        elif self.type == 'entityref':
            return f'&{self.name};'
        elif self.type == 'charref':
            return f'&#{self.name};'
        return self.__repr__()

    def _rewrite_attrs(self, attr):
        return attr.replace('\xA0', '&nbsp;').replace('&#x27;', "'")


class MarkupTag(Markup):
    @property
    def classes(self):
        try:
            return self._classes
        except AttributeError:
            pass
        classes_text = self.getattr('class')
        rv = [] if classes_text is None else REGEX_ASCII_WHITESPACES.split(classes_text)
        self._classes = rv
        return rv

    def getattr(self, attr, default=None):
        """This works as HTML5 spec that a boolean attribute returns ''.
        """
        for k, v in self.attrs:
            if k == attr:
                return v or ''
        return default


class HTMLParser(html.parser.HTMLParser):
    """HTML parser extended from native to support some HTML5 behaviors.
    """
    # include escapable raw text elements
    CDATA_CONTENT_ELEMENTS = ('script', 'style', 'xmp', 'textarea', 'title')

    def __init__(self, convert_charrefs=False, is_xhtml=False):
        super().__init__(convert_charrefs=convert_charrefs)
        self._is_xhtml = is_xhtml

        self._rv = []
        self._stack = []

    def error(self, message):
        raise Error(message)

    def parse_html_declaration(self, i):
        """Hook parent class to retrieve source text of special declaration
        """
        self.__last_added_decl_markup = None
        endpos = super().parse_html_declaration(i)
        if self.__last_added_decl_markup is not None:
            self.__last_added_decl_markup.src = self.rawdata[i:endpos]
        return endpos

    def parse_endtag(self, i):
        """Hook parent class to retrieve source text of endtag
        """
        self.__last_added_endtag_markup = None
        endpos = super().parse_endtag(i)
        if self.__last_added_endtag_markup is not None:
            self.__last_added_endtag_markup.src = self.rawdata[i:endpos]
        return endpos

    def handle_pi(self, data):
        self._process(Markup(
            is_xhtml=self._is_xhtml,
            type='pi',
            data=data,
        ))

    def handle_decl(self, decl):
        self._process(Markup(
            is_xhtml=self._is_xhtml,
            type='decl',
            data=decl,
        ))

    def handle_comment(self, data):
        markup = Markup(
            is_xhtml=self._is_xhtml,
            type='comment',
            data=data,
        )
        self.__last_added_decl_markup = markup
        self._process(markup)

    def handle_starttag(self, tag, attrs):
        self._process(MarkupTag(
            is_xhtml=self._is_xhtml,
            type='starttag',
            tag=tag,
            attrs=attrs,
            src=self.get_starttag_text(),
        ))

    def handle_startendtag(self, tag, attrs):
        self._process(MarkupTag(
            is_xhtml=self._is_xhtml,
            type='starttag',
            tag=tag,
            attrs=attrs,
            src=self.get_starttag_text(),
            is_self_end=True,
        ))

    def handle_endtag(self, tag):
        markup = MarkupTag(
            is_xhtml=self._is_xhtml,
            type='endtag',
            tag=tag,
        )
        self.__last_added_endtag_markup = markup
        self._process(markup)

    def handle_data(self, data):
        self._process(Markup(
            is_xhtml=self._is_xhtml,
            type='data',
            data=data,
            convert_charrefs=self.convert_charrefs,
            is_cdata=bool(self.cdata_elem),
        ))

    def handle_entityref(self, name):
        self._process(Markup(
            is_xhtml=self._is_xhtml,
            type='entityref',
            name=name,
        ))

    def handle_charref(self, name):
        self._process(Markup(
            is_xhtml=self._is_xhtml,
            type='charref',
            name=name,
        ))

    def unknown_decl(self, data):
        markup = Markup(
            is_xhtml=self._is_xhtml,
            type='decl',
            data=data,
            is_unknown=True,
        )
        self.__last_added_decl_markup = markup
        self._process(markup)

    def close(self):
        super().close()

        for _ in reversed(range(0, len(self._stack))):
            starttag = self._stack.pop()

            # create a hidden starttag for the unmatched starting tag
            endtag = MarkupTag(
                type='endtag',
                tag=starttag.tag,
                starttag=starttag,
                hidden=True,
            )
            starttag.endtag = endtag
            self._rv.append(endtag)

    def _process(self, markup):
        """Process according to the current parser status

        @TODO: further implementation of HTML5 standard to prevent a potential
               error due to not well-formed HTML
               ref: https://html.spec.whatwg.org/multipage/parsing.html
        """
        self._process_token(markup)

    def _process_token(self, markup):
        getattr(self, f'_process_token_{markup.type}')(markup)

    def _process_token_pi(self, markup):
        self._rv.append(markup)

    def _process_token_decl(self, markup):
        self._rv.append(markup)

    def _process_token_comment(self, markup):
        self._rv.append(markup)

    def _process_token_starttag(self, markup):
        self._rv.append(markup)

        require_endtag = False

        if self._is_xhtml or self._in_html_foreign_element():
            if markup.is_self_end:
                require_endtag = True

        elif markup.tag in VOID_ELEMENTS:
            require_endtag = True

        if require_endtag:
            # create a hidden endtag and don't push stack
            endtag = MarkupTag(
                is_xhtml=self._is_xhtml,
                type='endtag',
                tag=markup.tag,
                starttag=markup,
                hidden=True,
            )
            markup.endtag = endtag
            self._rv.append(endtag)
            return

        self._stack.append(markup)

    def _process_token_endtag(self, markup):
        for _ in reversed(range(0, len(self._stack))):
            starttag = self._stack.pop()
            if starttag.tag == markup.tag:
                starttag.endtag = markup
                markup.starttag = starttag
                break

            # create a hidden endtag for the implicitly closed tag
            endtag = MarkupTag(
                is_xhtml=self._is_xhtml,
                type='endtag',
                tag=starttag.tag,
                starttag=starttag,
                hidden=True,
            )
            starttag.endtag = endtag
            self._rv.append(endtag)
        else:
            # create a hidden starttag for the unmatched ending tag
            starttag = MarkupTag(
                is_xhtml=self._is_xhtml,
                type='starttag',
                tag=markup.tag,
                attrs=[],
                endtag=markup,
                hidden=True,
            )
            markup.starttag = starttag
            self._rv.append(starttag)

        self._rv.append(markup)

    def _process_token_data(self, markup):
        self._rv.append(markup)

    def _process_token_entityref(self, markup):
        self._rv.append(markup)

    def _process_token_charref(self, markup):
        self._rv.append(markup)

    def _ignore_markup(self, markup):
        markup.ignored = True
        self._rv.append(markup)

    def _in_html_foreign_element(self):
        for markup in self._stack:
            if markup.tag in FOREIGN_ELEMENTS:
                return True
        return False


class HtmlRewriter:
    """The base class that handles HTML rewriting for a specific path.
    """
    def __init__(self, file=None, *,
                 doc_url=None, base_url=None, url_chain=set(),  # noqa: B006
                 is_xhtml=None, encoding=None,
                 parser=HTMLParser):
        """Initialize the class and bind associated information.

        Args:
            file: path of the associated file.
            doc_url: overriding URL path of the document.
            base_url: overriding URL resolving path.
            is_xhtml: whether this file is xhtml.
            encoding: the encoding for reading a file. None for autodetection
                (and self.encoding will be auto reset on reading).
        """
        self.parser = parser
        self.file = file
        self.is_xhtml = is_xhtml
        self.encoding = encoding
        self.doc_url = doc_url
        self.base_url = base_url
        self.url_chain = url_chain.copy()

        if file:
            if is_xhtml is None:
                self.is_xhtml = util.is_xhtml(file) or util.is_svg(file)

            if doc_url is None:
                self.doc_url = urljoin('file:///', pathname2url(file))

        if self.doc_url:
            if base_url is None:
                self.base_url = self.doc_url

            self.url_chain.add(self.doc_url)

        self.changed = False

    def run(self, *args, **kwargs):
        """Common rewriting case when an associated file is provided.

        Can be overwritten by a subclass.
        """
        if not self.file:
            raise RuntimeError('Associated file not set.')

        self.changed = False

        markups = self.load(self.file)
        markups = self.rewrite(markups, *args, **kwargs)

        if self.changed:
            with open(self.file, 'w', encoding=self.encoding, newline='') as fh:
                for markup in markups:
                    if not markup.hidden:
                        fh.write(str(markup))

    def load(self, file):
        """Load a file and return parsed markups.

        May reset self.encoding.

        Args:
            file: str, path-like, or file-like bytes object
        Raises:
            OSError: failed to read the file
        """
        try:
            fh = open(file, 'rb')
        except TypeError:
            fh = file

        try:
            if not self.encoding:
                self.encoding = util.get_html_charset(fh)
                fh.seek(0)

            p = self.parser(is_xhtml=self.is_xhtml)
            for s in codecs.iterdecode(fh, self.encoding):
                if not s:
                    break
                p.feed(s)
            p.close()

            return p._rv
        finally:
            if fh != file:
                fh.close()

    def loads(self, text):
        """Load a string and return parsed markups.
        """
        p = self.parser(is_xhtml=self.is_xhtml)
        p.feed(text)
        p.close()

        return p._rv

    def rewrite(self, markups):
        """Main rewriter.

        Defaults to do nothing. Should be overwritten by a subclass.
        """
        return markups

    def find(self, markups, filter, start=0, endtag=None):
        return next(self.iterfind(markups, filter, start, endtag), None)

    def iterfind(self, markups, filter, start=0, endtag=None):
        i = start
        while True:
            try:
                markup = markups[i]
            except IndexError:
                break

            if filter(markup):
                yield i

            if endtag is not None:
                if markup == endtag:
                    break

            i += 1



"""The WGSI application.
"""
import functools
import json
import os
import time
import traceback
from collections import namedtuple
from contextlib import nullcontext
from secrets import token_urlsafe
from urllib.parse import quote, unquote, urljoin, urlsplit, urlunsplit
from zlib import adler32

import commonmark
import flask
import jinja2
from flask import (
    Response,
    abort,
    current_app,
    redirect,
    render_template,
    request,
)
from werkzeug.datastructures import WWWAuthenticate
from werkzeug.exceptions import HTTPException
from werkzeug.http import (
    dump_options_header,
    http_date,
    is_resource_modified,
    parse_options_header,
)
from werkzeug.local import LocalProxy
from werkzeug.middleware.proxy_fix import ProxyFix
from werkzeug.utils import cached_property

from . import WSB_CONFIG, WSB_DIR, WSB_EXTENSION_MIN_VERSION, __version__, util
from ._polyfill import mimetypes, zipfile
from .scrapbook import cache as wsb_cache
from .scrapbook import check as wsb_check
from .scrapbook import host as wsb_host
from .util.fs import (
    ZIP_SUBPATH_DIR,
    ZIP_SUBPATH_DIR_IMPLICIT,
    ZIP_SUBPATH_DIR_ROOT,
    ZIP_SUBPATH_FILE,
)

# see: https://url.spec.whatwg.org/#percent-encoded-bytes
quote_path = functools.partial(quote, safe=":/[]@!$&'()*+,;=")
quote_path.__doc__ = 'Escape reserved chars for the path part of a URL.'

bp = flask.Blueprint('default', __name__)
host = LocalProxy(lambda: current_app.config['WEBSCRAPBOOK_HOST'])


def static_url(path):
    return f'{quote_path(request.script_root)}/{quote_path(path)}?a=static'


def apply_csp(response):
    """Apply CSP, mostly for a static resource.
    """
    if host.config['app']['content_security_policy'] == 'strict':
        response.headers.set('Content-Security-Policy', "connect-src 'none'; form-action 'none';")


def static_file(filename, mimetype=None):
    """Output the specified file to the client.

    Args:
        filename: absolute path of the file to output.
    """
    if not os.path.isfile(filename):
        abort(404)
    response = flask.send_file(filename, conditional=True, mimetype=mimetype)
    response.headers.set('Accept-Ranges', 'bytes')
    response.headers.set('Cache-Control', 'no-cache')
    apply_csp(response)
    return response


def zip_static_file(zh, subpath, mimetype=None):
    """Output the specified file in a ZIP to the client.

    Args:
        zh: an opened zipfile.ZipFile
        subpath: str or zipfile.ZipInfo
    """
    if not isinstance(subpath, zipfile.ZipInfo):
        try:
            info = zh.getinfo(subpath)
        except KeyError:
            abort(404)
    else:
        info = subpath

    fh = zh.open(info, 'r')

    lm = util.fs.zip_timestamp(info)
    last_modified = http_date(lm)

    etag = '%s-%s-%s' % (
        lm,
        info.file_size,
        adler32(info.filename.encode('utf-8')) & 0xFFFFFFFF,
    )

    headers = {
        'Accept-Ranges': 'bytes',
        'Cache-Control': 'no-cache',
        'Last-Modified': last_modified,
        'ETag': etag,
    }

    response = Response(fh, headers=headers, mimetype=mimetype)
    response.make_conditional(request.environ, accept_ranges=True, complete_length=info.file_size)
    apply_csp(response)
    return response


def stream_template(template_name, **context):
    current_app.update_template_context(context)
    t = current_app.jinja_env.get_template(template_name)
    return t.stream(context)


def http_response(body='', status=None, headers=None, format=None):
    """Handle formatted response.

    ref: https://jsonapi.org
    """
    if not format:
        mimetype = None

    # expect body to be a JSON-serializable object
    elif format == 'json':
        mimetype = 'application/json'

        body = {
            'success': True,
            'data': body,
        }

        body = json.dumps(body, ensure_ascii=False)

    # expect body to be a generator of text (mostly JSON) data
    elif format == 'sse':
        mimetype = 'text/event-stream'

        def wrapper(gen):
            try:
                for data in gen:
                    yield 'data: ' + data + '\n\n'
            except Exception:
                traceback.print_exc()
                err = {'error': {'message': 'Internal Server Error'}}
                yield 'data: ' + json.dumps(err, ensure_ascii=False) + '\n\n'

            yield 'event: complete' + '\n'
            yield 'data: ' + '\n\n'

        body = wrapper(body)

    else:
        abort(400, f'Output format "{format}" is not supported.')

    return Response(body, status, headers, mimetype=mimetype)


def get_localpath(path):
    """Convert a request path to local filesystem path.
    """
    # Don't use os.path.join as it doesn't concatenate if path looks like
    # absolute (e.g. "/path/to/foo" on POSIX or "X:/foo" on Windows), which
    # can cause a security issue.
    return os.path.normpath(host.chroot + os.sep + path)


def get_breadcrumbs(localpaths, base='', topname='.'):
    """Generate (label, subpath, sep, is_last) tuples.
    """
    base = base.rstrip('/') + '/'
    paths = list(localpaths)
    paths[0] = paths[0].strip('/')

    if not paths[0]:
        yield (topname, base, '/', True)
        return

    yield (topname, base, '/', False)

    # handle zip root, which is something like /archive.zip!/
    is_zip_root = False
    if paths[-1] == '':
        paths.pop()
        is_zip_root = True

    paths_max = len(paths) - 1
    pathlist = []
    for path_idx, path in enumerate(paths):
        pathlist.append([])
        parts = path.split('/')
        parts_max = len(parts) - 1
        for part_idx, part in enumerate(parts):
            pathlist[-1].append(part)
            subpath = '!/'.join('/'.join(p) for p in pathlist)
            sep = '!/' if part_idx == parts_max and (path_idx < paths_max or is_zip_root) else '/'
            is_last = path_idx == paths_max and part_idx == parts_max
            yield (part, base + subpath + sep, sep, is_last)


def is_local_access():
    """Determine if the client is in same device.
    """
    server_host = request.host.partition(':')[0]
    client_host = request.remote_addr
    return util.is_localhost(server_host) or util.is_localhost(client_host) or server_host == client_host


def get_permission(auth_info, auth_config):
    """Calculate effective permission from provided auth info and config.
    """
    auth = auth_info or {}
    user = auth.get('username') or ''
    pw = auth.get('password') or ''

    for _, entry in auth_config.items():
        entry_user = entry.get('user', '')
        if user != entry_user:
            continue

        entry_pw = entry.get('pw', '')
        entry_pw_salt = entry.get('pw_salt', '')
        entry_pw_type = entry.get('pw_type', '')
        if util.encrypt(pw, entry_pw_salt, entry_pw_type) != entry_pw:
            continue

        entry_permission = entry.get('permission', 'all')
        return entry_permission

    return ''


def verify_authorization(perm, action):
    """Check if authorized or not.
    """
    if perm == 'all':
        return True

    if perm == 'read':
        return action not in {
            'token', 'lock', 'unlock',
            'mkdir', 'mkzip', 'save', 'delete', 'move', 'copy',
            'backup', 'unbackup', 'cache', 'check',
        }

    if perm == 'view':
        return action in {'view', 'info', 'source', 'download', 'static', 'unknown'}

    return False


def handle_directory_listing(localpaths, zh=None, redirect_slash=True, format=None):
    """List contents in a directory.

    Args:
        localpaths: a CPath
        zh: an opened zipfile.ZipFile object for faster reading
    """
    # ensure directory has trailing '/'
    if redirect_slash and not request.path.endswith('/'):
        parts = urlsplit(request.url)
        new_url = urlunsplit((
            parts.scheme,
            parts.netloc,
            quote_path(unquote(parts.path)) + '/',
            parts.query,
            parts.fragment,
        ))
        return redirect(new_url)

    # prepare index
    if len(localpaths) > 1:
        # support 304 if zip not modified
        stats = os.stat(localpaths[0])
        last_modified = http_date(stats.st_mtime)
        etag = '%s-%s-%s' % (
            stats.st_mtime,
            stats.st_size,
            adler32(localpaths[0].encode('utf-8')) & 0xFFFFFFFF,
        )

        if not is_resource_modified(request.environ, etag=etag, last_modified=last_modified):
            return http_response(status=304, format=format)

        headers = {
            'Cache-Control': 'no-cache',
            'Last-Modified': last_modified,
            'ETag': etag,
        }

        with nullcontext(zh) if zh else util.fs.open_archive_path(localpaths) as zh:
            subentries = zip_listdir(zh, localpaths[-1])

    else:
        # disallow cache to reflect any content file change
        stats = os.stat(localpaths[0])
        headers = {
            'Cache-Control': 'no-store',
            'Last-Modified': http_date(stats.st_mtime),
        }

        subentries = listdir(localpaths[0])

    if format == 'sse':
        def gen():
            for entry in subentries:
                data = entry._asdict()
                yield json.dumps(data, ensure_ascii=False)

        return http_response(gen(), headers=headers, format=format)

    if format == 'json':
        data = [e._asdict() for e in subentries]
        return http_response(data, headers=headers, format=format)

    body = render_template('index.html',
                           sitename=host.name,
                           is_local=is_local_access(),
                           base=request.script_root,
                           path=request.path,
                           pathparts=request.paths,
                           subentries=subentries,
                           )
    return http_response(body, headers=headers)


def handle_archive_viewing(localpaths, mimetype):
    """Handle direct visit of HTZ/MAFF file.

    Args:
        localpaths: a CPath
    """
    def list_maff_pages(pages):
        """List available web pages in a MAFF file.
        """
        return render_template('maff_index.html',
                               sitename=host.name,
                               is_local=is_local_access(),
                               base=request.script_root,
                               path=request.path,
                               pages=pages,
                               )

    if mimetype == 'application/html+zip':
        subpath = 'index.html'
    else:
        if len(localpaths) > 1:
            with util.fs.open_archive_path(localpaths) as zh:
                with zh.open(localpaths[-1]) as zh1:
                    pages = util.get_maff_pages(zh1)
        else:
            pages = util.get_maff_pages(localpaths[-1])

        if len(pages) > 1:
            # multiple index files
            return list_maff_pages(pages)

        if len(pages) == 0:
            # no valid index file found
            return list_maff_pages([])

        subpath = pages[0].indexfilename

    parts = urlsplit(request.url)
    new_url = urlunsplit((
        parts.scheme,
        parts.netloc,
        quote_path(unquote(parts.path)) + '!/' + quote_path(subpath),
        parts.query,
        parts.fragment,
    ))
    return redirect(new_url)


def handle_markdown_output(localpaths, zh=None):
    """Output processed markdown.

    Args:
        localpaths: a CPath
        zh: an opened zipfile.ZipFile object for faster reading
    """
    if len(localpaths) > 1:
        if zh:
            context = nullcontext(zh)
        else:
            context = util.fs.open_archive_path(localpaths)
    else:
        context = nullcontext(None)

    with context as zh:
        # calculate last-modified time and etag
        if zh:
            info = zh.getinfo(localpaths[-1])
            lm = util.fs.zip_timestamp(info)
            last_modified = http_date(lm)

            etag = '%s-%s-%s' % (
                lm,
                info.file_size,
                adler32(info.filename.encode('utf-8')) & 0xFFFFFFFF,
            )
        else:
            stats = os.stat(localpaths[0])
            last_modified = http_date(stats.st_mtime)
            etag = '%s-%s-%s' % (
                stats.st_mtime,
                stats.st_size,
                adler32(localpaths[0].encode('utf-8')) & 0xFFFFFFFF,
            )

        if not is_resource_modified(request.environ, etag=etag, last_modified=last_modified):
            return http_response(status=304)

        headers = {
            'Cache-Control': 'no-cache',
            'Last-Modified': last_modified,
            'ETag': etag,
        }

        # prepare content
        if zh:
            with zh.open(info) as fh:
                body = fh.read().decode('UTF-8')
        else:
            with open(localpaths[0], 'r', encoding='UTF-8') as fh:
                body = fh.read()

    body = render_template('markdown.html',
                           sitename=host.name,
                           is_local=is_local_access(),
                           base=request.script_root,
                           path=request.path,
                           pathparts=request.paths,
                           content=commonmark.commonmark(body),
                           )

    response = http_response(body, headers=headers)
    apply_csp(response)
    return response


class Request(flask.Request):
    """Subclassed Request object for more useful properties.
    """
    @cached_property
    def paths(self):
        """Like request.path, but with ZIP subpaths resolved."""
        return util.fs.CPath.resolve(self.path, get_localpath)

    @cached_property
    def localpath(self):
        """Corresponding filesystem path of the requested path."""
        return get_localpath(self.path)

    @cached_property
    def localpaths(self):
        """Like localpath, but with ZIP subpaths resolved."""
        return util.fs.CPath(
            get_localpath(self.paths[0]),
            *self.paths[1:],
        )

    @cached_property
    def localrealpath(self):
        """Like localpath, but with symlinks resolved."""
        return os.path.realpath(self.localpath)

    @cached_property
    def localmimetype(self):
        """Mimetype of the requested path."""
        mimetype, _ = mimetypes.guess_type(self.localrealpath)
        return mimetype

    @cached_property
    def action(self):
        """Shortcut of the requested action."""
        rv = request.values.get('a', default='view')
        rv = request.values.get('action', default=rv)
        if not globals().get(f'action_{rv}'):
            rv = 'unknown'
        return rv

    @cached_property
    def format(self):
        """Shortcut of the requested format."""
        rv = request.values.get('f')
        rv = request.values.get('format', default=rv)
        return rv


def handle_action_token(func):
    """A decorator function that validates token.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        token = request.values.get('token') or ''

        if not host.token_validate(token):
            abort(400, 'Invalid access token.')

        host.token_delete(token)

        return func(*args, **kwargs)

    return wrapper


def handle_action_advanced(func):
    """A decorator function that helps handling an advanced command.

    - Verify POST method.
    - Provide a default return value.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        format = request.format

        if request.method != 'POST':
            abort(405, valid_methods=['POST'])

        rv = func(*args, **kwargs)

        if rv is not None:
            return rv

        if format:
            return http_response('Command run successfully.', format=format)

        return http_response(status=204)

    return wrapper


def handle_action_writing(func):
    """A decorator function that helps handling a writing action.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if os.path.abspath(request.localpath) == host.chroot:
            abort(403, 'Unable to operate the root directory.')

        return func(*args, **kwargs)

    return wrapper


def action_unknown():
    """Default handler for an undefined action"""
    abort(400, 'Action not supported.')


def action_view():
    """Show the content of a file or list a directory.

    If formatted, show information of the file or directory.
    """
    # info for other output formats
    if request.format:
        return action_info()

    localpaths = request.localpaths
    mimetype = request.localmimetype

    if len(localpaths) > 1:
        with util.fs.open_archive_path(localpaths) as zh:
            # List directory only when URL suffixed with "/", as it's not a
            # common operation, and it's costy to check for directory existence
            # in a ZIP.
            if request.path.endswith('/'):
                try:
                    return handle_directory_listing(localpaths, zh, redirect_slash=False)
                except ZipDirNotFoundError:
                    abort(404)

            try:
                info = zh.getinfo(localpaths[-1])
            except KeyError:
                # File does not exist.
                abort(404)
            else:
                # view archive file
                if mimetype in ('application/html+zip', 'application/x-maff'):
                    return handle_archive_viewing(localpaths, mimetype)

                # view markdown
                if mimetype == 'text/markdown':
                    return handle_markdown_output(localpaths, zh)

                # convert meta refresh to 302 redirect
                if localpaths[-1].lower().endswith('.htm'):
                    with zh.open(info) as fh:
                        target = util.get_meta_refresh(fh).target

                    if target is not None:
                        # Keep several chars as javascript encodeURI do,
                        # plus "%" as target may have already been escaped.
                        parts = urlsplit(urljoin(request.url, quote(target, ";,/?:@&=+$-_.!~*'()#%")))
                        new_url = urlunsplit((
                            parts.scheme,
                            parts.netloc,
                            quote_path(unquote(parts.path)),
                            parts.query,
                            parts.fragment,
                        ))
                        return redirect(new_url)

                # show static file for other cases
                response = zip_static_file(zh, localpaths[-1], mimetype=mimetype)
    else:
        localpath = localpaths[0]

        # handle directory
        if os.path.isdir(localpath):
            return handle_directory_listing(localpaths)

        # handle file
        elif os.path.isfile(localpath):
            # view archive file
            if mimetype in ('application/html+zip', 'application/x-maff'):
                return handle_archive_viewing(localpaths, mimetype)

            # view markdown
            if mimetype == 'text/markdown':
                return handle_markdown_output(localpaths)

            # convert meta refresh to 302 redirect
            if request.localrealpath.lower().endswith('.htm'):
                target = util.get_meta_refresh(localpath).target

                if target is not None:
                    # Keep several chars as javascript encodeURI do,
                    # plus "%" as target may have already been escaped.
                    parts = urlsplit(urljoin(request.url, quote(target, ";,/?:@&=+$-_.!~*'()#%")))
                    new_url = urlunsplit((
                        parts.scheme,
                        parts.netloc,
                        quote_path(unquote(parts.path)),
                        parts.query,
                        parts.fragment,
                    ))
                    return redirect(new_url)

            # show static file for other cases
            response = static_file(localpath, mimetype=mimetype)

        else:
            abort(404)

    # don't include charset
    m, p = parse_options_header(response.headers.get('Content-Type'))
    try:
        del p['charset']
    except KeyError:
        pass
    response.headers.set('Content-Type', dump_options_header(m, p))

    return response


def action_source():
    """Show file content as plain text."""
    if request.format:
        abort(400, 'Action not supported.')

    localpaths = request.localpaths

    if len(localpaths) > 1:
        with util.fs.open_archive_path(localpaths) as zh:
            response = zip_static_file(zh, localpaths[-1])
    else:
        response = static_file(localpaths[0])

    # show as inline plain text
    # @TODO: Chromium (80) seems to ignore header mimetype for certain types
    #        like image and zip
    encoding = request.values.get('e', 'utf-8')
    encoding = request.values.get('encoding', default=encoding)
    response.headers.set('Content-Type', 'text/plain; charset=' + quote(encoding))
    response.headers.set('Content-Disposition', 'inline')

    return response


def action_download():
    """Download a file or directory."""
    if request.format:
        abort(400, 'Action not supported.')

    localpaths = request.localpaths
    filter = request.values.getlist('i')

    if len(localpaths) > 1:
        with util.fs.open_archive_path(localpaths) as zh:
            cur = util.fs.zip_check_subpath(zh, localpaths[-1], allow_invalid=True)
            if cur in (ZIP_SUBPATH_DIR, ZIP_SUBPATH_DIR_IMPLICIT, ZIP_SUBPATH_DIR_ROOT):
                filename = (os.path.basename(localpaths[-1]) or os.path.basename(localpaths[-2])) + '.zip'
                mimetype, _ = mimetypes.guess_type(filename)
                response = None
            elif cur == ZIP_SUBPATH_FILE:
                filename = os.path.basename(request.localrealpath)
                response = zip_static_file(zh, localpaths[-1], mimetype=request.localmimetype)
            else:
                abort(404)

        if not response:
            def gen():
                with util.fs.open_archive_path(localpaths) as zh:
                    yield from util.fs.zip_copy(zh, localpaths[-1], None, '', filter)

            response = Response(gen(), mimetype=mimetype)
            response.headers.set('Cache-Control', 'no-store')
    else:
        if os.path.isdir(localpaths[0]):
            filename = os.path.basename(request.localrealpath) + '.zip'
            mimetype, _ = mimetypes.guess_type(filename)
            gen = util.fs.zip_compress(None, localpaths[0], '', filter)
            response = Response(gen, mimetype=mimetype)
            response.headers.set('Cache-Control', 'no-store')
        else:
            filename = os.path.basename(request.localrealpath)
            response = static_file(localpaths[0])

    filename = quote_path(filename)
    response.headers.set('Content-Disposition',
                         f'''attachment; filename*=UTF-8''{filename}; filename="{filename}"''')
    return response


def action_info():
    """Show information of a path."""
    format = request.format

    if not format:
        abort(400, 'Action not supported.')

    localpaths = request.localpaths
    mimetype = request.localmimetype

    if len(localpaths) > 1:
        with util.fs.open_archive_path(localpaths) as zh:
            info = zip_file_info(zh, localpaths[-1], check_implicit_dir=True)
    else:
        info = file_info(localpaths[0])

    data = info._asdict()
    data['mime'] = mimetype
    return http_response(data, format=format)


def action_list():
    """List entries in a directory."""
    format = request.format

    if not format:
        abort(400, 'Action not supported.')

    localpaths = request.localpaths

    if len(localpaths) > 1:
        try:
            return handle_directory_listing(localpaths, redirect_slash=False, format=format)
        except ZipDirNotFoundError:
            abort(404, 'Directory does not exist.')

    if os.path.isdir(localpaths[0]):
        return handle_directory_listing(localpaths, redirect_slash=False, format=format)

    abort(404, 'Directory does not exist.')


def action_static():
    """Show a static file of the current theme."""
    format = request.format

    if format:
        abort(400, 'Action not supported.')

    filepath = request.path.strip('/')
    file = host.get_static_file(filepath)
    if file:
        return static_file(file)

    abort(404)


def action_edit():
    """Simple text editor for a file."""
    format = request.format

    if format:
        abort(400, 'Action not supported.')

    localpaths = request.localpaths
    localpath = localpaths[0]

    if os.path.lexists(localpath) and not os.path.isfile(localpath):
        abort(400, 'Found a non-file here.')

    if len(localpaths) > 1:
        with util.fs.open_archive_path(localpaths) as zh:
            try:
                info = zh.getinfo(localpaths[-1])
            except KeyError:
                body = b''
            else:
                body = zh.read(info)
    else:
        try:
            with open(localpath, 'rb') as fh:
                body = fh.read()
        except FileNotFoundError:
            body = b''

    encoding = request.values.get('e')
    encoding = request.values.get('encoding', default=encoding)

    try:
        body = body.decode(encoding or 'UTF-8')
    except (LookupError, UnicodeDecodeError):
        encoding = 'ISO-8859-1'
        body = body.decode(encoding)

    body = render_template('edit.html',
                           sitename=host.name,
                           is_local=is_local_access(),
                           base=request.script_root,
                           path=request.path,
                           body=body,
                           encoding=encoding,
                           )

    return http_response(body, format=format)


def action_editx():
    """HTML editor for a file."""
    format = request.format

    if format:
        abort(400, 'Action not supported.')

    localpaths = request.localpaths
    localpath = localpaths[0]

    if os.path.lexists(localpath) and not os.path.isfile(localpath):
        abort(400, 'Found a non-file here.')

    if request.localmimetype not in ('text/html', 'application/xhtml+xml'):
        abort(400, 'This is not an HTML file.')

    if len(localpaths) > 1:
        with util.fs.open_archive_path(localpaths) as zh:
            try:
                zh.getinfo(localpaths[-1])
            except KeyError:
                abort(404)
    else:
        if not os.path.lexists(localpath):
            abort(404)

    body = render_template('editx.html',
                           sitename=host.name,
                           is_local=is_local_access(),
                           base=request.script_root,
                           path=request.path,
                           )

    return http_response(body, format=format)


def action_exec():
    """Launch a file or directory."""
    format = request.format

    if not is_local_access():
        abort(400, 'Command can only run on local device.')

    localpath = request.localpath

    if not os.path.lexists(localpath):
        abort(404, 'File does not exist.')

    util.fs.launch(localpath)

    if format:
        return http_response('Command run successfully.', format=format)

    return http_response(status=204)


def action_browse():
    """Open a file or directory in the file browser."""
    format = request.format

    if not is_local_access():
        abort(400, 'Command can only run on local device.')

    localpath = request.localpath

    if not os.path.lexists(localpath):
        abort(404, 'File does not exist.')

    util.fs.view_in_explorer(localpath)

    if format:
        return http_response('Command run successfully.', format=format)

    return http_response(status=204)


def action_config():
    """Show server config."""
    format = request.format

    if not format:
        abort(400, 'Action not supported.')

    data = host.config.dump_object()

    # filter values for better security
    data = {k: v for k, v in data.items() if k in ('app', 'book')}
    data['app'] = {k: v for k, v in data['app'].items() if k in ('name', 'theme', 'locale')}

    # expose backup_dir if it's web accessible
    if os.path.normcase(os.path.join(host.backup_dir, '')).startswith(os.path.normcase(os.path.join(host.chroot, ''))):
        data['app']['backup_dir'] = host.backup_dir[len(os.path.join(host.chroot, '')):].replace('\\', '/')

    # add and rewrite values for client to better know the server
    data['app']['base'] = request.script_root
    data['app']['is_local'] = is_local_access()
    data['VERSION'] = __version__
    data['WSB_DIR'] = WSB_DIR
    data['WSB_CONFIG'] = WSB_CONFIG
    data['WSB_EXTENSION_MIN_VERSION'] = WSB_EXTENSION_MIN_VERSION

    return http_response(data, format=format)


def action_token():
    """Acquire a token and return its name."""
    format = request.format

    # require POST method
    if request.method != 'POST':
        abort(405, valid_methods=['POST'])

    return http_response(host.token_acquire(), format=format)


@handle_action_advanced
@handle_action_token
def action_lock():
    """Acquire a lock for the given name.

    URL params:
    - id: for persisting and extending previous lock.
    - chkt: timeout to wait for the lock.
    """
    # verify name
    name = request.values.get('name')
    if name is None:
        abort(400, 'Lock name is not specified.')

    id = request.values.get('id')

    timeout = request.values.get('chkt', 5, type=float)

    try:
        lock = host.get_lock(name, persist=id)
    except wsb_host.LockPersistError:
        abort(400, f'Unable to persist lock "{name}".')

    if id:
        try:
            lock.extend()
        except wsb_host.LockExtendNotFoundError:
            # Lock file gone in this short interval. Try acquire.
            pass
        except wsb_host.LockExtendError:
            abort(500, f'Unable to extend lock "{name}".')
        else:
            return http_response(lock.id, format=request.format)

    try:
        lock.acquire(timeout=timeout)
    except wsb_host.LockTimeoutError:
        abort(503, f'Unable to acquire lock "{name}".', retry_after=60)
    except wsb_host.LockRegenerateError:
        abort(500, f'Unable to regenerate stale lock "{name}".')
    except wsb_host.LockGenerateError:
        abort(500, f'Unable to create lock "{name}".')

    return http_response(lock.id, format=request.format)


@handle_action_advanced
@handle_action_token
def action_unlock():
    """Release a lock for the given name."""
    # verify name
    name = request.values.get('name')
    if name is None:
        abort(400, 'Lock name is not specified.')

    # verify ID
    id = request.values.get('id')
    if id is None:
        abort(400, 'Lock ID is not specified.')

    try:
        lock = host.get_lock(name, persist=id)
    except wsb_host.LockPersistError:
        abort(400, f'Unable to persist lock "{name}".')

    try:
        lock.release()
    except wsb_host.LockReleaseNotFoundError:
        pass
    except wsb_host.LockReleaseError:
        abort(500, f'Unable to remove lock "{name}".')


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_mkdir():
    """Create a directory."""
    localpaths = request.localpaths

    try:
        util.fs.mkdir(localpaths)
    except util.fs.FSEntryExistsError:
        abort(400, 'Found something here.')
    except util.fs.FSBadParentError:
        abort(400, 'Parent directory is not available.')
    except Exception:
        traceback.print_exc()
        abort(500, 'Unable to create a directory here.')


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_mkzip():
    """Create a zip file."""
    localpaths = request.localpaths

    try:
        util.fs.mkzip(localpaths)
    except util.fs.FSIsADirectoryError:
        abort(400, 'Found a non-file here.')
    except util.fs.FSEntryExistsError:
        abort(400, 'Found something here.')
    except util.fs.FSBadParentError:
        abort(400, 'Parent directory is not available.')
    except Exception:
        traceback.print_exc()
        abort(500, 'Unable to write to this ZIP file.')


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_save():
    """Write a file with provided text or uploaded stream."""
    localpaths = request.localpaths

    file = request.files.get('upload')
    if file is not None:
        src = file.stream
    else:
        src = request.values.get('text', '').encode('ISO-8859-1')

    try:
        util.fs.save(localpaths, src)
    except util.fs.FSIsADirectoryError:
        abort(400, 'Found a non-file here.')
    except util.fs.FSEntryExistsError:
        abort(400, 'Found something here.')
    except util.fs.FSBadParentError:
        abort(400, 'Parent directory is not available.')
    except Exception:
        traceback.print_exc()
        abort(500, 'Unable to write to this ZIP file.')


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_delete():
    """Delete a file or directory."""
    localpaths = request.localpaths

    try:
        util.fs.delete(localpaths)
    except util.fs.FSEntryNotFoundError:
        abort(404, 'Entry does not exist.')
    except Exception:
        traceback.print_exc()
        abort(500, 'Unable to delete this entry.')


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_move():
    """Move a file or directory."""
    localpaths = request.localpaths

    target = request.values.get('target')
    if target is None:
        abort(400, 'Target is not specified.')
    targetpaths = util.fs.CPath.resolve(target, get_localpath).path
    targetpaths[0] = get_localpath(targetpaths[0])

    try:
        util.fs.move(localpaths, targetpaths)
    except util.fs.FSEntryExistsError:
        abort(400, 'Target already exists.')
    except util.fs.FSMoveInsideError:
        abort(400, 'Unable to move into self.')
    except util.fs.FSMoveAcrossZipError:
        abort(400, 'Unable to move across a zip.')
    except util.fs.FSEntryNotFoundError:
        abort(404, 'Source does not exist.')
    except Exception:
        traceback.print_exc()
        abort(500, 'Unable to move to the target.')


@handle_action_advanced
@handle_action_token
@handle_action_writing
def action_copy():
    """Copy a file or directory."""
    localpaths = request.localpaths

    target = request.values.get('target')
    if target is None:
        abort(400, 'Target is not specified.')
    targetpaths = util.fs.CPath.resolve(target, get_localpath).path
    targetpaths[0] = get_localpath(targetpaths[0])

    try:
        util.fs.copy(localpaths, targetpaths)
    except util.fs.FSEntryExistsError:
        abort(400, 'Target already exists.')
    except util.fs.FSEntryNotFoundError:
        abort(404, 'Source does not exist.')
    except util.fs.FSPartialError:
        abort(500, 'Fail to copy some files.')
    except Exception:
        traceback.print_exc()
        abort(500, 'Unable to copy to the target.')


@handle_action_advanced
@handle_action_token
def action_backup():
    """Bakup file or directory."""
    format = request.format

    if not format:
        abort(400, 'Action not supported.')

    localpaths = request.localpaths

    if len(localpaths) > 1:
        abort(400, 'Unable to backup inside a zip file.')

    ts = request.values.get('ts') or util.datetime_to_id()
    note = request.values.get('note')
    move = request.values.get('move', default=False, type=bool)

    host.init_backup(ts, note=note)
    try:
        host.auto_backup(localpaths[0], move=move)
        return http_response(os.path.basename(host._backup_dir), format=request.format)
    finally:
        host.init_backup(False)


@handle_action_advanced
@handle_action_token
def action_unbackup():
    """Remove a backup."""
    format = request.format

    if not format:
        abort(400, 'Action not supported.')

    ts = request.values.get('ts') or util.datetime_to_id()
    note = request.values.get('note')

    host.init_backup(ts, note=note)
    try:
        host.unbackup(host._backup_dir)
        return http_response(os.path.basename(host._backup_dir), format=request.format)
    finally:
        host.init_backup(False)


@handle_action_token
def action_cache():
    """Invoke the cacher."""
    format = request.format

    kwargs = {
        'book_ids': request.values.getlist('book'),
        'item_ids': request.values.getlist('item'),
        'no_lock': request.values.get('no_lock', default=False, type=bool),
        'no_backup': request.values.get('no_backup', default=False, type=bool),
        'fulltext': request.values.get('fulltext', default=False, type=bool),
        'inclusive_frames': request.values.get('inclusive_frames', default=False, type=bool),
        'recreate': request.values.get('recreate', default=False, type=bool),
        'static_site': request.values.get('static_site', default=False, type=bool),
        'static_index': request.values.get('static_index', default=False, type=bool),
        'rss_root': request.values.get('rss_root'),
        'rss_item_count': request.values.get('rss_item_count', default=50, type=int),
        'locale': request.values.get('locale'),
    }

    headers = {
        'Cache-Control': 'no-store',
    }
    root = host.root
    config = host.config

    if format == 'sse':
        def gen():
            for info in wsb_cache.generate(root, config=config, **kwargs):
                data = {
                    'type': info.type,
                    'msg': info.msg,
                }

                yield json.dumps(data, ensure_ascii=False)

        return http_response(gen(), headers=headers, format=format)

    elif format:
        abort(400, 'Action not supported.')

    def gen():
        yield from wsb_cache.generate(root, config=config, **kwargs)

    stream = stream_template('cli.html',
                             title='Indexing...',
                             messages=gen(),
                             debug=False,
                             )

    return Response(stream, headers=headers)


@handle_action_token
def action_check():
    """Invoke the checker."""
    format = request.format

    kwargs = {
        'book_ids': request.values.getlist('book'),
        'no_lock': request.values.get('no_lock', default=False, type=bool),
        'no_backup': request.values.get('no_backup', default=False, type=bool),
        'resolve_invalid_id': request.values.get('resolve_invalid_id', default=False, type=bool),
        'resolve_missing_index': request.values.get('resolve_missing_index', default=False, type=bool),
        'resolve_missing_index_file': request.values.get('resolve_missing_index_file', default=False, type=bool),
        'resolve_missing_date': request.values.get('resolve_missing_date', default=False, type=bool),
        'resolve_older_mtime': request.values.get('resolve_older_mtime', default=False, type=bool),
        'resolve_toc_unreachable': request.values.get('resolve_toc_unreachable', default=False, type=bool),
        'resolve_toc_invalid': request.values.get('resolve_toc_invalid', default=False, type=bool),
        'resolve_toc_empty_subtree': request.values.get('resolve_toc_empty_subtree', default=False, type=bool),
        'resolve_unindexed_files': request.values.get('resolve_unindexed_files', default=False, type=bool),
        'resolve_absolute_icon': request.values.get('resolve_absolute_icon', default=False, type=bool),
        'resolve_unused_icon': request.values.get('resolve_unused_icon', default=False, type=bool),
    }

    headers = {
        'Cache-Control': 'no-store',
    }
    root = host.root
    config = host.config

    if format == 'sse':
        def gen():
            for info in wsb_check.run(root, config=config, **kwargs):
                data = {
                    'type': info.type,
                    'msg': info.msg,
                }

                yield json.dumps(data, ensure_ascii=False)

        return http_response(gen(), headers=headers, format=format)

    elif format:
        abort(400, 'Action not supported.')

    def gen():
        yield from wsb_check.run(root, config=config, **kwargs)

    stream = stream_template('cli.html',
                             title='Indexing...',
                             messages=gen(),
                             debug=False,
                             )

    return Response(stream, headers=headers)


@bp.before_request
def handle_before_request():
    # handle authorization
    try:
        auth_config = host.config['auth']
    except KeyError:
        # auth not required
        return

    perm = get_permission(request.authorization, auth_config)
    if not verify_authorization(perm, request.action):
        auth = WWWAuthenticate()
        auth.set_basic(host.config['app']['name'])
        abort(401, 'You are not authorized.', www_authenticate=auth)


@bp.route('/', methods=['GET', 'HEAD', 'POST'])
@bp.route('/<path:filepath>', methods=['GET', 'HEAD', 'POST'])
def handle_request(filepath=''):
    """Handle an HTTP request (HEAD, GET, POST).
    """
    try:
        handler = globals()[f'action_{request.action}']
        return handler()
    except PermissionError:
        abort(403)


@bp.after_request
def handle_after_request(response):
    # forbid a privileged page to be framed
    if host.config['app']['content_security_policy'] == 'strict':
        if 'Content-Security-Policy' not in response.headers:
            response.headers.set('Content-Security-Policy', "frame-ancestors 'none';")
            response.headers.set('X-Frame-Options', 'deny')

    return response


@bp.errorhandler(HTTPException)
def handle_error(exc):
    """Handle formatted error if requested by client.
    """
    if request.format == 'json':
        response = exc.get_response()
        response.data = json.dumps({
            'error': {
                'status': exc.code,
                'message': exc.description,
            },
        })
        response.content_type = 'application/json'
        return response

    return exc


class WebHost(wsb_host.Host):
    """Extended Host class that also handles HTTP server related things.

    - Token handling: security token validation to avoid CSRF attack.
    """
    TOKEN_PURGE_INTERVAL = 3600  # in seconds
    TOKEN_DEFAULT_EXPIRY = 1800  # in seconds

    def __init__(self, root, config=None):
        super().__init__(root, config=config)

        # token handling
        self.tokens = os.path.join(self.root, WSB_DIR, 'server', 'tokens')
        self.token_last_purge = 0

    def token_acquire(self, now=None):
        if now is None:
            now = int(time.time())

        self.token_check_delete_expire(now)

        token = token_urlsafe()
        token_file = os.path.join(self.tokens, token)
        while os.path.lexists(token_file):
            token = token_urlsafe()
            token_file = os.path.join(self.tokens, token)

        os.makedirs(os.path.dirname(token_file), exist_ok=True)
        with open(token_file, 'w', encoding='UTF-8') as fh:
            fh.write(str(now + self.TOKEN_DEFAULT_EXPIRY))

        return token

    def token_validate(self, token, now=None):
        if now is None:
            now = int(time.time())

        token_file = os.path.join(self.tokens, token)

        try:
            with open(token_file, 'r', encoding='UTF-8') as fh:
                expire = int(fh.read())
        except (FileNotFoundError, IsADirectoryError):
            return False

        if now >= expire:
            os.remove(token_file)
            return False

        return True

    def token_delete(self, token):
        token_file = os.path.join(self.tokens, token)

        try:
            os.remove(token_file)
        except OSError:
            pass

    def token_delete_expire(self, now=None):
        if now is None:
            now = int(time.time())

        try:
            token_files = os.scandir(self.tokens)
        except FileNotFoundError:
            pass
        else:
            for token_file in token_files:
                try:
                    with open(token_file, 'r', encoding='UTF-8') as fh:
                        expire = int(fh.read())
                except (OSError, ValueError):
                    continue
                if now >= expire:
                    os.remove(token_file)

    def token_check_delete_expire(self, now=None):
        if now is None:
            now = int(time.time())

        if now >= self.token_last_purge + self.TOKEN_PURGE_INTERVAL:
            self.token_last_purge = now
            self.token_delete_expire(now)


def make_app(root='.', config=None):
    _host = WebHost(root, config=config)

    # main app instance
    app = flask.Flask(__name__, instance_path=_host.chroot)
    app.register_blueprint(bp)
    app.request_class = Request
    app.config['WEBSCRAPBOOK_HOST'] = _host

    xheaders = {
        'x_for': _host.config['app']['allowed_x_for'],
        'x_proto': _host.config['app']['allowed_x_proto'],
        'x_host': _host.config['app']['allowed_x_host'],
        'x_port': _host.config['app']['allowed_x_port'],
        'x_prefix': _host.config['app']['allowed_x_prefix'],
    }

    if any(v for v in xheaders.values()):
        app.wsgi_app = ProxyFix(app.wsgi_app, **xheaders)

    app.jinja_loader = jinja2.FileSystemLoader(_host.templates)
    app.jinja_env.globals.update({
        'os': os,
        'time': time,
        'get_breadcrumbs': get_breadcrumbs,
        'format_filesize': functools.partial(util.format_filesize, space='\xA0'),
        'quote_path': quote_path,
        'static_url': static_url,
        'i18n': _host.get_i18n(_host.config['app']['locale']),
    })

    return app


#########################################################################
# Filesystem helpers
#########################################################################

FileInfo = namedtuple('FileInfo', ('name', 'type', 'size', 'last_modified'))


def file_info(file, base=None):
    """Read basic file information.

    Args:
        file: path of the file
        base: path that the result filename is based under
    """
    if base is None:
        name = os.path.basename(file)
    else:
        base = os.path.join(base, '')
        if not file.startswith(base):
            raise ValueError('file not under base')

        name = file[len(base):]
        if os.sep != '/':
            name = name.replace(os.sep, '/')

    try:
        statinfo = os.lstat(file)
    except OSError:
        # unexpected error when getting stat info
        statinfo = None
        size = None
        last_modified = None
    else:
        size = statinfo.st_size
        last_modified = statinfo.st_mtime

    if not os.path.lexists(file):
        type = None
    elif os.path.islink(file) or util.fs.isjunction(file):
        type = 'link'
    elif os.path.isdir(file):
        type = 'dir'
    elif os.path.isfile(file):
        type = 'file'
    else:
        type = 'unknown'

    if type != 'file':
        size = None

    return FileInfo(name=name, type=type, size=size, last_modified=last_modified)


def listdir(base, recursive=False):
    """Generates FileInfo(s) and omit invalid entries.
    """
    if not recursive:
        with os.scandir(base) as entries:
            for entry in entries:
                info = file_info(entry.path)
                if info.type is None:
                    continue
                yield info

    else:
        for root, dirs, files in os.walk(base):
            for dir in dirs:
                file = os.path.join(root, dir)
                info = file_info(file, base)
                if info.type is None:
                    continue
                yield info
            for file in files:
                file = os.path.join(root, file)
                info = file_info(file, base)
                if info.type is None:
                    continue
                yield info


#########################################################################
# ZIP helpers
#########################################################################

class ZipDirNotFoundError(Exception):
    pass


def zip_file_info(zip, subpath, base=None, check_implicit_dir=False):
    """Read basic file information from ZIP.

    Args:
        zip: path, file-like object, or zipfile.ZipFile
        subpath: 'dir' and 'dir/' are both supported
        base: path that the result filename is based under,
            'dir' and 'dir/' are both supported
    """
    subpath = subpath.rstrip('/')
    if base is None:
        name = os.path.basename(subpath)
    else:
        base = base.rstrip('/')
        base = base + ('/' if base else '')
        if not subpath.startswith(base):
            raise ValueError('subpath not under base')

        name = subpath[len(base):]

    with nullcontext(zip) if isinstance(zip, zipfile.ZipFile) else zipfile.ZipFile(zip) as zh:
        try:
            info = zh.getinfo(subpath)
        except KeyError:
            pass
        else:
            return FileInfo(
                name=name, type='file',
                size=info.file_size,
                last_modified=util.fs.zip_timestamp(info),
            )

        try:
            info = zh.getinfo(subpath + '/')
        except KeyError:
            pass
        else:
            return FileInfo(
                name=name, type='dir', size=None,
                last_modified=util.fs.zip_timestamp(info),
            )

        if check_implicit_dir:
            base = subpath + ('/' if subpath else '')
            for entry in zh.namelist():
                if entry.startswith(base):
                    return FileInfo(name=name, type='dir', size=None, last_modified=None)

    return FileInfo(name=name, type=None, size=None, last_modified=None)


def zip_listdir(zip, subpath, recursive=False):
    """Generates FileInfo(s) and omit invalid entries.

    Raise ZipDirNotFoundError if subpath does not exist.

    NOTE: It is possible that entry mydir/ does not exist while mydir/foo.bar
    exists. Check for matching subentries to make sure whether the implicit
    directory exists.

    Args:
        zip: path, file-like object, or zipfile.ZipFile
        subpath: the subpath in the ZIP, with or without trailing slash
    """
    base = subpath.rstrip('/')
    if base:
        base += '/'
    base_len = len(base)
    dir_exist = not base
    entries = {}

    with nullcontext(zip) if isinstance(zip, zipfile.ZipFile) else zipfile.ZipFile(zip) as zh:
        for filename in zh.namelist():
            if not filename.startswith(base):
                continue

            if filename == base:
                dir_exist = True
                continue

            entry = filename[base_len:]
            if not recursive:
                entry, _, _ = entry.partition('/')
                entries.setdefault(entry, True)
            else:
                parts = entry.rstrip('/').split('/')
                for i in range(0, len(parts)):
                    entry = '/'.join(parts[0:i + 1])
                    entries.setdefault(entry, True)

        if not entries and not dir_exist:
            raise ZipDirNotFoundError(f'Directory "{base}" does not exist in the zip.')

        for entry in entries:
            info = zip_file_info(zh, base + entry, base)

            if info.type is None:
                yield FileInfo(name=entry, type='dir', size=None, last_modified=None)
            else:
                yield info



"""Command line interface of WebScrapBook toolkit.
"""
import argparse
import os
import shutil
import sys
import traceback
from getpass import getpass
from inspect import getdoc
from time import time_ns

from . import (
    WSB_CONFIG,
    WSB_DIR,
    WSB_USER_CONFIG,
    __version__,
    config,
    server,
    util,
)


def log(*args):
    print(*args)


def die(*args):
    print('Error:', *args, file=sys.stderr)
    sys.exit(1)


def get_umask():
    """Get configured umask.
    """
    umask = os.umask(0)
    os.umask(umask)
    return umask


def fcopy(fsrc, fdst):
    """Copy a script file to target

    - Auto generate ancestor directories.
    - Use universal linefeed.
    - Set last modified time to current time.
    """
    os.makedirs(os.path.dirname(os.path.abspath(fdst)), exist_ok=True)
    with open(fsrc, 'r', encoding='UTF-8') as fr:
        with open(fdst, 'w', encoding='UTF-8') as fw:
            for line in fr:
                fw.write(line)


def cmd_serve(args):
    """Serve the root directory forever. Shutdown via Ctrl+C or another killing
    technique.

    By default the local browser will be launched to view the served (hosted) site.
    This behavior can be changed using the `server.browse` config.

    Note that this built-in server is designed only for local hosting, or remote
    hosting for personal or few people usage. For an opened world wide web hosting,
    a more specialized server should be used."""
    server.serve(**args)


def cmd_config(args):
    """Show, generate, or edit the config.

    Display the current config when used with no arguments.

    Run `wsb help config` for details about config.
    """
    if args['book']:
        fdst = os.path.normpath(os.path.join(args['root'], WSB_DIR, WSB_CONFIG))
        fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', 'config.ini'))
        if not os.path.isfile(fdst):
            log(f'Generating "{fdst}"...')
            try:
                fcopy(fsrc, fdst)
            except OSError:
                die(f'Unable to generate {fdst}.')

        if args['edit']:
            try:
                util.fs.launch(fdst)
            except OSError:
                pass

        if args['all']:
            fdst = os.path.normpath(os.path.join(args['root'], WSB_DIR, 'serve.py'))
            fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', 'serve.py'))
            if not os.path.isfile(fdst):
                log(f'Generating "{fdst}"...')
                try:
                    fcopy(fsrc, fdst)
                    os.chmod(fdst, os.stat(fdst).st_mode | (0o111 & ~get_umask()))
                except OSError:
                    die(f'Unable to generate {fdst}.')

            fdst = os.path.normpath(os.path.join(args['root'], WSB_DIR, 'app.py'))
            fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', 'app.py'))
            if not os.path.isfile(fdst):
                log(f'Generating "{fdst}"...')
                try:
                    fcopy(fsrc, fdst)
                    os.chmod(fdst, os.stat(fdst).st_mode | (0o111 & ~get_umask()))
                except OSError:
                    die(f'Unable to generate {fdst}.')

    elif args['user']:
        fdst = WSB_USER_CONFIG
        fsrc = os.path.normpath(os.path.join(__file__, '..', 'resources', 'config.ini'))
        if not os.path.isfile(fdst):
            log(f'Generating "{fdst}"...')
            try:
                fcopy(fsrc, fdst)
            except OSError:
                die(f'Unable to generate {fdst}.')

        if args['edit']:
            try:
                util.fs.launch(fdst)
            except OSError:
                pass

    elif args['edit']:
        die('Use --edit in combine with --book or --user.')

    elif args['all']:
        die('Use --all in combine with --book.')

    elif args['name']:
        config.load(args['root'])
        value = config.getname(args['name'])

        if value is None:
            die(f"""Config entry "{args['name']}" does not exist""")

        print(value)

    else:
        config.load(args['root'])
        config.dump(sys.stdout)


def cmd_encrypt(args):
    """Generate an encrypted password string.

    Primilarly to be used in auth config.
    """
    if args['password'] is None:
        pw1 = getpass('Enter a password: ')
        pw2 = getpass('Confirm the password: ')

        if pw1 != pw2:
            die('Entered passwords do not match.')

        args['password'] = pw1

    print(util.encrypt(args['password'], salt=args['salt'], method=args['method']))


def cmd_cache(args):
    """Generate (or update) fulltext cache and/or static site pages.
    """
    kwargs = args.copy()
    debug = kwargs.pop('debug')

    from .scrapbook import cache
    for info in cache.generate(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_export(args):
    """Export data items into archive files (*.wsba).
    """
    kwargs = args.copy()
    debug = kwargs.pop('debug')

    from .scrapbook import exporter
    for info in exporter.run(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_import(args):
    """Import data items from archive files (*.wsba).

    To faithfully recover original tree structure, import items using the same
    order as how they have been exported.
    """
    kwargs = args.copy()
    debug = kwargs.pop('debug')

    from .scrapbook import importer
    for info in importer.run(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_check(args):
    """Integrity check and fix for scrapbook data.

    (TOC = table of contents)
    """
    kwargs = args.copy()
    debug = kwargs.pop('debug')

    from .scrapbook import check
    for info in check.run(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_convert(args):
    """Convert data between different formats.

    Do not perform any data operation for the input and/or output directory
    during the conversion process (and related backend server(s) should be shut
    down in prior) to prevent a potential conversion error.
    """
    kwargs = args.copy()
    kwargs.pop('root')
    mode = kwargs.pop('mode')
    force = kwargs.pop('force')
    debug = kwargs.pop('debug')

    import importlib
    conv = importlib.import_module(f'.scrapbook.convert.{mode}', __package__)

    # validate input and output directory
    input = args['input']
    input = os.path.realpath(input)

    if not os.path.isdir(input):
        die(f'''Input directory not available: "{input}"''')

    output = args['output']
    if output is not None:
        output = os.path.realpath(output)

        if os.path.normcase(output) == os.path.normcase(input):
            die("""Unable to output to the input directory""")

        if os.path.normcase(output).startswith(os.path.normcase(os.path.join(input, ''))):
            die("""Unable to output to a descendant of the input directory""")

        if os.path.normcase(input).startswith(os.path.normcase(os.path.join(output, ''))):
            die("""Unable to output to an ancestor of the input directory""")

        if not os.path.lexists(output):
            pass
        elif not os.path.isdir(output):
            die(f'''Output directory not available: "{output}"''')
        else:
            if force:
                # using os.rmtree() frequently cause an error on Windows
                with os.scandir(output) as dirs:
                    for entry in dirs:
                        try:
                            shutil.rmtree(entry)
                        except NotADirectoryError:
                            os.remove(entry)
            else:
                with os.scandir(output) as dirs:
                    if next(dirs, None):
                        die(f'''Output directory not empty: "{output}"''')

    for info in conv.run(**kwargs):
        if info.type != 'debug' or debug:
            log(f'{info.type.upper()}: {info.msg}')


def cmd_help(args):
    """Show detailed information about certain topics.
    """
    root = os.path.join(os.path.dirname(__file__), 'resources')

    if args['topic'] == 'config':
        file = os.path.join(root, 'config.md')
    elif args['topic'] == 'themes':
        file = os.path.join(root, 'themes.md')
    elif args['topic'] == 'mimetypes':
        file = os.path.join(root, 'mimetypes.md')

    if file:
        with open(file, 'r', encoding='UTF-8') as fh:
            text = fh.read()
        print(text)


def cmd_view(args):
    """View archive file(s) in the browser.

    Supported formats: *.htz, *.maff
    """
    config.load(args['root'])
    view_archive_files(args['files'])


def view_archive_files(files):
    """View archive file(s) in the browser.

    Set default application of MAFF/HTZ archive files to this command to open
    them in the browser directly.
    """
    import tempfile
    import webbrowser
    from urllib.request import pathname2url

    from ._polyfill import mimetypes, zipfile

    cache_prefix = config['browser']['cache_prefix']
    cache_expire = config['browser']['cache_expire'] * 10 ** 9
    use_jar = config['browser']['use_jar']
    browser = webbrowser.get(config['browser']['command'] or None)

    temp_dir = tempfile.gettempdir()
    urls = []

    for file in dict.fromkeys(os.path.normcase(os.path.abspath(file)) for file in files):
        mime, _ = mimetypes.guess_type(file)
        if mime not in ('application/html+zip', 'application/x-maff'):
            continue

        if use_jar:
            base_url = 'jar:file:' + pathname2url(file) + '!/'
            if mime == 'application/html+zip':
                urls.append(base_url + 'index.html')
            elif mime == 'application/x-maff':
                urls.extend(base_url + f.indexfilename for f in util.get_maff_pages(file))
            continue

        # extract zip contents to dest_dir if not done yet
        hash = util.checksum(file)
        dest_prefix = cache_prefix + hash + '_'
        with os.scandir(temp_dir) as entries:
            for entry in entries:
                if not entry.name.startswith(dest_prefix):
                    continue

                dest_dir = entry.path

                # update atime
                atime = time_ns()
                stat = os.stat(entry)
                os.utime(entry, ns=(atime, stat.st_mtime_ns))
                break
            else:
                dest_dir = tempfile.mkdtemp(prefix=dest_prefix)
                with zipfile.ZipFile(file) as zh:
                    zh.extractall(dest_dir)

        # get URL of every index page
        base_url = 'file:' + pathname2url(dest_dir) + '/'
        if mime == 'application/html+zip':
            urls.append(base_url + 'index.html')
        elif mime == 'application/x-maff':
            urls.extend(base_url + f.indexfilename for f in util.get_maff_pages(file))

    # open pages in the browser
    for url in urls:
        browser.open(url)

    # remove stale caches
    if not use_jar:
        t = time_ns()
        with os.scandir(temp_dir) as entries:
            for entry in entries:
                if not entry.name.startswith(cache_prefix):
                    continue

                atime = os.stat(entry).st_atime_ns
                if t <= atime + cache_expire:
                    continue

                # cache may be created by another user and undeletable
                try:
                    shutil.rmtree(entry)
                except OSError:
                    traceback.print_exc()


def view():
    """CLI entry point for viewing archive files.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(view_archive_files))
    parser.add_argument(
        'files', nargs='+',
        help="""files to view.""")
    args = vars(parser.parse_args())
    view_archive_files(args['files'])


def parse_args(argv=None):
    # Improve program name when executed through python -m
    # NOTE: We don't expect a bad command name such as having a space.
    if os.path.basename(sys.argv[0]) == '__main__.py':
        prog = f'{os.path.basename(sys.executable)} -m webscrapbook'
    else:
        prog = None

    parser = argparse.ArgumentParser(prog=prog, description=__doc__)
    parser.add_argument(
        '--version', action='version', version=f'{__package__} {__version__}',
        help="""show version information and exit""")
    parser.add_argument(
        '--root', default='.',
        help="""root directory to manipulate (default: current working directory)""")
    subparsers = parser.add_subparsers(
        metavar='COMMAND',
        help="""the sub-command to run. Get usage help with e.g. %(prog)s config -h""")

    # subcommand: serve
    parser_serve = subparsers.add_parser(
        'serve', aliases=['s'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_serve),
        help="""serve the root directory""")
    parser_serve.add_argument(
        '--browse', default=None, action='store_true',
        help="""launch the browser to visit the served directory""")
    parser_serve.add_argument(
        '--no-browse', dest='browse', action='store_false',
        help="""do not launch the browser""")
    parser_serve.set_defaults(func=cmd_serve)

    # subcommand: config
    parser_config = subparsers.add_parser(
        'config', aliases=['c'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_config),
        help="""show, generate, or edit the config""")
    parser_config.set_defaults(func=cmd_config)
    parser_config.add_argument(
        'name', nargs='?',
        help="""show value of the given config name (in the form of <section>[.<subsection>].<key>)""")
    parser_config.add_argument(
        '-b', '--book', default=False, action='store_true',
        help="""generate book (host) config file""")
    parser_config.add_argument(
        '-u', '--user', default=False, action='store_true',
        help="""generate user config file""")
    parser_config.add_argument(
        '-a', '--all', default=False, action='store_true',
        help="""generate more assistant files (with --book)""")
    parser_config.add_argument(
        '-e', '--edit', default=False, action='store_true',
        help="""edit the config file (with --book or --user)""")

    # subcommand: encrypt
    parser_encrypt = subparsers.add_parser(
        'encrypt', aliases=['e'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_encrypt),
        help="""generate an encrypted password""")
    parser_encrypt.set_defaults(func=cmd_encrypt)
    parser_encrypt.add_argument(
        '-p', '--password', nargs='?', default=None, action='store',
        help="""the password to encrypt. Skip to provide via an interactive prompt.""")
    parser_encrypt.add_argument(
        '-m', '--method', default='sha1', action='store',
        help="""the encrypt method to use, which is one of: plain, md5, sha1,
sha224, sha256, sha384, sha512, sha3_224, sha3_256, sha3_384, and sha3_512
(default: %(default)s)""")
    parser_encrypt.add_argument(
        '-s', '--salt', default='', action='store',
        help="""the salt to add during encryption.""")

    # subcommand: cache
    parser_cache = subparsers.add_parser(
        'cache', aliases=['a'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_cache),
        help="""update fulltext cache and/or static site pages""")
    parser_cache.set_defaults(func=cmd_cache)
    parser_cache.add_argument(
        'book_ids', metavar='book', nargs='*', action='store',
        help="""the book ID(s) to generate cache. (default: all books)""")
    parser_cache.add_argument(
        '--item', dest='item_ids',
        metavar='ID', action='store', default=None, nargs='+',
        help="""the items ID(s) to generate cache (default: all)""")
    parser_cache.add_argument(
        '--fulltext', default=True, action='store_true',
        help="""generate fulltext cache. (default)""")
    parser_cache.add_argument(
        '--no-fulltext', dest='fulltext', action='store_false',
        help="""inverse of --fulltext""")
    parser_cache.add_argument(
        '--inclusive-frames', default=True, action='store_true',
        help="""cache frame content as part of the main page (default). It's
recommended to recreate fulltext cache when changing this option to prevent
inconsistency.""")
    parser_cache.add_argument(
        '--no-inclusive-frames', dest='inclusive_frames', action='store_false',
        help="""inverse of --inclusive-frames""")
    parser_cache.add_argument(
        '--recreate', dest='recreate', default=False, action='store_true',
        help="""ignore current fulltext cache and generate again""")
    parser_cache.add_argument(
        '--no-recreate', dest='recreate', action='store_false',
        help="""inverse of --recreate (default)""")
    parser_cache.add_argument(
        '--static-site', default=False, action='store_true',
        help="""generate static site pages""")
    parser_cache.add_argument(
        '--no-static-site', dest='static_site', action='store_false',
        help="""inverse of --static-site (default)""")
    parser_cache.add_argument(
        '--static-index', default=False, action='store_true',
        help="""generate static index.html page""")
    parser_cache.add_argument(
        '--no-static-index', dest='static_index', action='store_false',
        help="""inverse of --static-index (default)""")
    parser_cache.add_argument(
        '--rss-root', metavar='ROOT_URL', action='store',
        help="""generate an RSS feed file for the book, using the specified root URL
        (usually corresponds to webscrapbook app root)""")
    parser_cache.add_argument(
        '--rss-item-count', default=50, type=int, action='store',
        help="""number of items the RSS feed should include (default: %(default)s)""")
    parser_cache.add_argument(
        '--locale', action='store',
        help="""locale for the generated pages (default: system locale)""")
    parser_cache.add_argument(
        '--backup', dest='no_backup', default=True, action='store_false',
        help="""backup changed files""")
    parser_cache.add_argument(
        '--no-backup', action='store_true',
        help="""do not backup changed files (default)""")
    parser_cache.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: check
    parser_check = subparsers.add_parser(
        'check', aliases=['k'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_check),
        help="""check and fix scrapbook data""")
    parser_check.set_defaults(func=cmd_check)
    parser_check.add_argument(
        'book_ids', metavar='book', nargs='*', action='store',
        help="""the book ID(s) to check. (default: all books)""")

    parser_check.add_argument(
        '-r', '--resolve', dest='resolve_all', default=False, action='store_true',
        help="""resolve all found issues (implies all --resolve-*)""")
    parser_check.add_argument(
        '--resolve-invalid-id', default=False, action='store_true',
        help="""remove items with invalid ID from metadata entries""")
    parser_check.add_argument(
        '--resolve-missing-index', default=False, action='store_true',
        help="""remove items with missing index property from metadata entries""")
    parser_check.add_argument(
        '--resolve-missing-index-file', default=False, action='store_true',
        help="""remove items with missing index file from metadata entries""")
    parser_check.add_argument(
        '--resolve-missing-date', default=False, action='store_true',
        help="""attempt to generate "create" and "modify" properties for items missing any of them""")
    parser_check.add_argument(
        '--resolve-older-mtime', default=False, action='store_true',
        help="""update "modify" property if it's older than last modified time of the index file""")
    parser_check.add_argument(
        '--resolve-toc-unreachable', default=False, action='store_true',
        help="""append items unreachable from TOC to the root tree""")
    parser_check.add_argument(
        '--resolve-toc-invalid', default=False, action='store_true',
        help="""remove invalid items from TOC""")
    parser_check.add_argument(
        '--resolve-toc-empty-subtree', default=False, action='store_true',
        help="""remove items with empty subtree from TOC""")
    parser_check.add_argument(
        '--resolve-unindexed-files', default=False, action='store_true',
        help="""attempt to import unindexed files to metadata and TOC""")
    parser_check.add_argument(
        '--resolve-absolute-icon', default=False, action='store_true',
        help="""cache "icon" property with absolute URL to local favicon directory""")
    parser_check.add_argument(
        '--resolve-unused-icon', default=False, action='store_true',
        help="""remove unused favicon caches""")

    parser_check.add_argument(
        '--no-backup', default=False, action='store_true',
        help="""do not backup changed files""")
    parser_check.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: export
    parser_export = subparsers.add_parser(
        'export', aliases=['x'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_export),
        help="""export data items into archive files (*.wsba)""")
    parser_export.set_defaults(func=cmd_export)
    parser_export.add_argument(
        'output', action='store',
        help="""the output directory""")
    parser_export.add_argument(
        '--book', dest='book_id', metavar='ID', default='', action='store',
        help="""the book ID to export. (default: "")""")
    parser_export.add_argument(
        '--item', dest='item_ids',
        metavar='ID', action='store', default=None, nargs='+',
        help="""the items ID(s) to export (default: all)""")
    parser_export.add_argument(
        '-r', '--recursive', default=False, action='store_true',
        help="""recursively include descendant items of the provided item ID(s)""")
    parser_export.add_argument(
        '-s', '--singleton', default=False, action='store_true',
        help="""export only the first occurrence for an item referenced many times""")
    parser_export.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: import
    parser_import = subparsers.add_parser(
        'import', aliases=['i'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_import),
        help="""import data items from archive files (*.wsba)""",
        epilog="""\
Available MODEs for --resolve-id-used:
  "skip": skip the import
  "replace": replace current metadata and data file(s)
  "new": import as a new item

Available placeholders for --filename PATTERN:
  "%%": a literal "%"
  "%ID%": item ID
  "%EID%": item export ID
  "%UUID%": a random UUID
  "%TITLE%": item title
  "%SOURCE%": item source URL
  "%CREATE%": item create time (allow SUBPATTERN)
  "%MODIFY%": item modify time (allow SUBPATTERN)
  "%EXPORT%": item export time (allow SUBPATTERN)

Datetime related ones can be written as %PATTERN:SUBPATTERN% for further
formatting with SUBPATTERN being one of these placeholders:
  "DATE": date (YYYY-MM-DD)
  "TIME": time (HH-MM-SS)
  "YEAR": year (YYYY)
  "MONTH": year (MM, 01-12)
  "DAY": day of month (DD, 01-31)
  "HOURS": hours (HH, 01-24)
  "MINUTES": minutes (MM, 00-59)
  "SECONDS": seconds (SS, 00-59)

All SUBPATTERNs can be prepended with "UTC_" for UTC time instead of local
time. For example, "%CREATE:UTC_DATE%".
""")
    parser_import.set_defaults(func=cmd_import)
    parser_import.add_argument(
        'files', metavar='file', action='store', nargs='+',
        help="""the file(s) to import in order. If a directory is provided, all
child files are imported in unicode filename order.""")
    parser_import.add_argument(
        '--book', dest='book_id', metavar='ID', default='', action='store',
        help="""the book ID to import into. (default: "")""")
    parser_import.add_argument(
        '--target', dest='target_id', metavar='ID',
        default='root', action='store',
        help="""the target item ID to insert the imported items under (default: "%(default)s")""")
    parser_import.add_argument(
        '--target-index', metavar='INDEX',
        type=int, action='store',
        help="""the index number (starting from 0) the imported items will be
inserted at (default: last)""")
    parser_import.add_argument(
        '--rebuild-folders', default=False, action='store_true',
        help="""insert imported items under the original parent, and
auto-generate parent folders if not found. (ignores --target and
--target-index)""")
    parser_import.add_argument(
        '--resolve-id-used', metavar='MODE',
        default='skip', action='store',
        choices={'skip', 'replace', 'new'},
        help="""what to do if an importing item ID already exists (default: "%(default)s")""")
    parser_import.add_argument(
        '--filename', dest='target_filename', metavar='PATTERN',
        default='%ID%', action='store',
        help="""formatter of the imported filename (default: "%(default)s")""")
    parser_import.add_argument(
        '--prune', default=False, action='store_true',
        help="""delete the archive file after successfully imported""")
    parser_import.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: convert
    parser_convert = subparsers.add_parser(
        'convert', aliases=['v'],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_convert),
        help="""convert scrapbook data between different formats""")
    parser_convert.set_defaults(func=cmd_convert)
    parser_convert_sub = parser_convert.add_subparsers(
        dest='mode', metavar='MODE',
        help="""the conversion mode. Get usage help with e.g. %(prog)s sb2wsb -h""")

    # -- migrate
    parser_convert_migrate = parser_convert_sub.add_parser(
        'migrate',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""\
Migrate a scrapbook to be compatible with the latest WebScrapBook version.
- WebScrapBook < 0.115: Migrate old shadow root loader and content data.
- WebScrapBook < 0.69: Migrate old canvas and shadow root loaders.

Also fix an incomplete conversion from legacy ScrapBook:
- Convert legacy annotations to be compatible with latest WebScrapBook.
- Convert legacy resources at chrome://scrapbook/skin/* to work in a browser
  without legacy ScrapBook add-on installed.
- Convert the index file of "postit" items for canonical wrapper and styling.
""",
        help="""migrate to latest WebScrapBook""")
    parser_convert_migrate.add_argument(
        'input', action='store',
        help="""the input directory""")
    parser_convert_migrate.add_argument(
        'output', action='store', nargs='?',
        help="""the output directory (default: in-place)""")
    parser_convert_migrate.add_argument(
        '--book', dest='book_ids', metavar='ID',
        nargs='+', action='store',
        help="""ID of the book(s) to convert (default: all books)""")
    parser_convert_migrate.add_argument(
        '--convert-legacy', default=True,
        action='store_true',
        help="""convert data files from legacy ScrapBook (default)""")
    parser_convert_migrate.add_argument(
        '--no-convert-legacy', dest='convert_legacy',
        action='store_false',
        help="""inverse of --convert-legacy""")
    parser_convert_migrate.add_argument(
        '--convert-v1', default=True,
        action='store_true',
        help="""convert data to latest WebScrapBook 1.* (default)""")
    parser_convert_migrate.add_argument(
        '--no-convert-v1', dest='convert_v1',
        action='store_false',
        help="""inverse of --convert-convert-v1""")
    parser_convert_migrate.add_argument(
        '--use-native-tags', default=False,
        action='store_true',
        help="""use native HTML tags for converted legacy ScrapBook annotations for better
compatibility with very old browsers (e.g. IE < 9), with the cost of increased possibility
to conflict with the web page stylesheets""")
    parser_convert_migrate.add_argument(
        '--no-use-native-tags', dest='use_native_tags',
        action='store_false',
        help="""inverse of --use-native-tags (default)""")
    parser_convert_migrate.add_argument(
        '--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_migrate.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # -- items
    parser_convert_items = parser_convert_sub.add_parser(
        'items',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Convert items in the scrapbook.""",
        help="""convert items in the scrapbook""",
        epilog="""\
Available FORMATs:
  "folder"       a folder with index.html and its resource files
  "htz"          a ZIP archive with index.html as entry
  "maff"         a ZIP archive with each page in an individial subfolder
  "single_file"  a single file with resources embedded in

* Conversion between "folder", "htz", and "maff" is mostly lostless (as long as
  no multi-page or arbitrary metadata is included in the MAFF file), while
  conversion between "single_file" and other formats may lose information
  permanently, such as filename and in-depth captured pages. Also note that
  some features are not supported for "single_file".

Available TYPEs:
  ""         a captured web page
  "site"     an in-depth captured web page set
  "file"     a captured or uploaded file
  "image"    an image file captured by legacy ScrapBook
  "note"     a note
  "postit"   a postit
  "combine"  a combined page generated by legacy ScrapBook
""")
    parser_convert_items.add_argument(
        'input', action='store',
        help="""the input directory""")
    parser_convert_items.add_argument(
        'output', action='store', nargs='?',
        help="""the output directory (default: in-place)""")
    parser_convert_items.add_argument(
        '--book', dest='book_ids', metavar='ID',
        nargs='+', action='store',
        help="""ID of the book(s) to convert (default: all books)""")
    parser_convert_items.add_argument(
        '--item', dest='item_ids', metavar='ID',
        nargs='+', action='store',
        help="""ID of the item(s) to convert (default: all items)""")
    parser_convert_items.add_argument(
        '--format', metavar='FORMAT', action='store',
        choices=['folder', 'htz', 'maff', 'single_file'],
        help="""file format to convert item(s) to (default: no conversion)""")
    parser_convert_items.add_argument(
        '--type', dest='types', metavar='TYPE', action='store', nargs='+',
        default=[''],
        choices=['', 'site', 'image', 'file', 'combine', 'note', 'postit', 'bookmark', 'folder', 'separator'],
        help="""item type(s) to convert (default: "")""")
    parser_convert_items.add_argument(
        '--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_items.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # -- sb2wsb
    parser_convert_sb2wsb = parser_convert_sub.add_parser(
        'sb2wsb',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""\
Convert a scrapbook from legacy ScrapBook to WebScrapBook.

The conversion is safe and lossless, with a few caveats:
- In-depth capture information is not converted and cannot be reused
  for a merge capture.
- Fulltext cache is not converted and requires a manual rebuild.

Known supported legacy scrapbook implementations:
- ScrapBook X (legacy Firefox Add-on)
- ScrapBook (legacy Firefox Add-on)
- ScrapBook Plus (legacy Firefox Add-on)
- ScrapBee (Firefox Quantum Add-on)""",
        help="""convert from legacy ScrapBook to WebScrapBook""")
    parser_convert_sb2wsb.add_argument(
        'input', action='store',
        help="""the input directory""")
    parser_convert_sb2wsb.add_argument(
        'output', action='store',
        help="""the output directory""")
    parser_convert_sb2wsb.add_argument(
        '--no-data-files', default=False, action='store_true',
        help="""do not convert data files (set this if there's something wrong
with the conversion, and run "wsb convert migrate" afterwards for advanced options)""")
    parser_convert_sb2wsb.add_argument(
        '--no-backup', default=False, action='store_true',
        help="""do not copy legacy ScrapBook files not needed by WebScrapBook""")
    parser_convert_sb2wsb.add_argument(
        '--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_sb2wsb.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # -- wsb2sb
    parser_convert_wsb2sb = parser_convert_sub.add_parser(
        'wsb2sb',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""\
Convert a scrapbook from WebScrapBook to legacy ScrapBook.

Fulltext cache is not converted and requires a manual rebuild.

Note that certain information may lose permanently, such as:
* items appended to multiple parents (preserve only the first occurence)
* items in the recycle bin
* inter-item links may break

Also note that compatibility validation of this tool is targeting ScrapBook X.
There may be minor compatibility issues if the output scrapbook is used by a
legacy ScrapBook implementation without features introduced by ScrapBook X,
such as:
* file with special or non-ASCII chars in filename
* container item whose type property is not "folder" """,
        help="""convert from WebScrapBook to legacy ScrapBook""")
    parser_convert_wsb2sb.add_argument(
        'input', action='store',
        help="""the input directory""")
    parser_convert_wsb2sb.add_argument(
        'output', action='store',
        help="""the output directory""")
    parser_convert_wsb2sb.add_argument(
        '--book', dest='book_id', metavar='ID',
        default='', action='store',
        help="""ID of the book to convert (default: "")""")
    parser_convert_wsb2sb.add_argument(
        '--no-data-files', default=False, action='store_true',
        help="""do not convert data files (set this if there's something wrong
for the conversion)""")
    parser_convert_wsb2sb.add_argument(
        '--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_wsb2sb.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # -- file2wsb
    parser_convert_file2wsb = parser_convert_sub.add_parser(
        'file2wsb',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""\
Convert hierarchical files to a WebScrapBook scrapbook.

This "imports" a folder with web page captures and possibly supporting folders
or sub-folders into structured scrapbook items. Metadata recorded by the
original capture tool, such as WebScrapBook, SingleFile, or native browser
saving, are translated into item metadata as much as possible. This is a lossy
conversion—the folder structure are changed and interlinkings may be broken.
""",
        help="""convert from hierarchical files to WebScrapBook""")
    parser_convert_file2wsb.add_argument(
        'input', action='store',
        help="""the input directory""")
    parser_convert_file2wsb.add_argument(
        'output', action='store',
        help="""the output directory""")
    parser_convert_file2wsb.add_argument(
        '--data-folder-suffix', dest='data_folder_suffixes',
        metavar='SUFFIX', default=None, action='store', nargs='*',
        help="""suffixes of the associated support folder (default: .files _files)""")
    parser_convert_file2wsb.add_argument(
        '--no-preserve-filename', default=False, action='store_true',
        help="""allow the converter to rename source files to reduce
folders and redirections in the output""")
    parser_convert_file2wsb.add_argument(
        '--ignore-ie-meta', default=False, action='store_true',
        help="""ignore metadata generated by built-in save of Internet
Explorer or a Chromium-based browser""")
    parser_convert_file2wsb.add_argument(
        '--ignore-singlefile-meta', default=False, action='store_true',
        help="""ignore metadata generated by SingleFile browser extension""")
    parser_convert_file2wsb.add_argument(
        '--ignore-savepagewe-meta', default=False, action='store_true',
        help="""ignore metadata generated by Save Page WE browser extension""")
    parser_convert_file2wsb.add_argument(
        '--ignore-maoxian-meta', default=False, action='store_true',
        help="""ignore metadata generated by MaoXian web clipper browser extension""")
    parser_convert_file2wsb.add_argument(
        '--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_file2wsb.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # -- wsb2file
    parser_convert_wsb2file = parser_convert_sub.add_parser(
        'wsb2file',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""\
Convert a WebScrapBook scrapbook to hierarchical files.

This turns the scrapbook items into structured physical folders and files to be
browsable with a file manager. Every output folder and file is prefixed with a
number (by default) to keep the original order in the tree. This is a lossy
conversion—item metadata are not preserved and interlinkings may be broken.

* For an item titled `MyItem` with index file `*/index.html`, its descendant
  items will be put under the `MyItem` subfolder, while its data files will be
  put under the `MyItem.htd` folder.
* A separator titled `MySep` will become the empty file `MySep.-`.
""",
        help="""convert from WebScrapBook to hierarchical files""")
    parser_convert_wsb2file.add_argument(
        'input', action='store',
        help="""the input directory""")
    parser_convert_wsb2file.add_argument(
        'output', action='store',
        help="""the output directory""")
    parser_convert_wsb2file.add_argument(
        '--book', dest='book_id', metavar='ID',
        default='', action='store',
        help="""ID of the book to convert (default: "")""")
    parser_convert_wsb2file.add_argument(
        '--no-prefix', dest='prefix', default=True, action='store_false',
        help="""don't prefix output files with position number.""")
    parser_convert_wsb2file.add_argument(
        '--force', default=False, action='store_true',
        help="""overwrite everything in the output directory""")
    parser_convert_wsb2file.add_argument(
        '--debug', default=False, action='store_true',
        help="""include debug output""")

    # subcommand: help
    parser_help = subparsers.add_parser(
        'help',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_help),
        help="""show detailed information about certain topics""",
        epilog="""\
Available TOPICs:
  "config"
  "themes"
  "mimetypes"
""")
    parser_help.set_defaults(func=cmd_help)
    parser_help.add_argument(
        'topic', metavar='TOPIC', default=None, action='store',
        choices=['config', 'themes', 'mimetypes'],
        help="""the topic for details""")

    # subcommand: view
    parser_view = subparsers.add_parser(
        'view',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=getdoc(cmd_view),
        help="""view archive file in the browser""")
    parser_view.set_defaults(func=cmd_view)
    parser_view.add_argument(
        'files', nargs='+',
        help="""files to view""")

    return parser.parse_args(argv)


def main():
    args = vars(parse_args())
    try:
        func = args.pop('func')
    except KeyError:
        parse_args(['-h'])
        return
    else:
        if func is cmd_convert and args['mode'] is None:
            parse_args(['convert', '-h'])
            return

        func(args)


if __name__ == '__main__':
    main()



bidi_dir = 'ltr'
html_lang = 'en'

index_title = 'Index of {path}'

data_table_header_directory = 'Directory'
data_table_header_name = 'Name'
data_table_header_last_modified = 'Last modified'
data_table_header_size = 'Size'

explorer_tooltip = 'View mode [{}]'
explorer_table = 'Table view'
explorer_gallery = 'Gallery view'
explorer_gallery2 = 'Gallery view (+media)'

tools_tooltip = 'Tools [{}]'
tools_preview_on = 'Enable preview'
tools_preview_off = 'Disable preview'
tools_select_all = 'Select all'
tools_deselect_all = 'Deselect all'
tools_expand_all = 'Expand all'
tools_filter = 'Filter'
tools_filter_clear = 'Clear filter'

command_tooltip = 'Commands [{}]'
command_mkdir = 'New folder'
command_mkzip = 'New zip'
command_mkfile = 'New file'
command_upload = 'Upload'
command_uploaddir = 'Upload folder'
command_source = 'View source'
command_download = 'Download'
command_exec = 'Run natively'
command_browse = 'Browse natively'
command_edit = 'Edit'
command_editx = 'Edit page'
command_move = 'Move'
command_copy = 'Copy'
command_link = 'Create link'
command_delete = 'Delete'

command_move_prompt = 'Move to the path or folder:'
command_move_prompt_multi = 'Move to the folder:'
command_copy_prompt = 'Copy to the path or folder:'
command_copy_prompt_multi = 'Copy to the folder:'
command_link_prompt = 'Create a link at the path or under the folder:'
command_link_prompt_multi = 'Create links under the folder:'

label_selected = '%count% selected'

previewer_toolbar_title = """\
Keybord shortcuts:
• [Esc]: close preview.
• [I] or [Space]: toggle infobar.
• [Enter]: open in new tab.
• [PageUp] or [Alt+←] or [←]: show previous entry.
• [PageDown] or [Alt+→] or [→]: show next entry.
• [Home]: show first entry.
• [End]: show last entry.
• [+] or [Alt+↑] or [↑]: zoom in.
• [-] or [Alt+↓] or [↓]: zoom out.
• [Ctrl+0]: zoom fit viewport.
• [Ctrl+1]: zoom between natural size and last zoom size.
• [Ctrl+←] or [Shift+←]: move left.
• [Ctrl+→] or [Shift+→]: move right.
• [Ctrl+↑] or [Shift+↑]: move up.
• [Ctrl+↓] or [Shift+↓]: move down.
"""
previewer_button_previous = 'Previous'
previewer_button_next = 'Next'
previewer_button_infobar = 'Toggle infobar'
previewer_button_close = 'Close preview'

edit_title = 'Edit {path}'
edit_error_no_javascript = 'Enable JavaScript or upgrade your browser to support JavaScript, so that the HTML editor can work.'
edit_warning_encoding = 'This file is decoded as {encoding}, and will be encoded as UTF-8 when saved.'

markdown_title = 'Page {path}'
maff_index_title = 'Pages of {path}'

button_save = 'SAVE'
button_save_tooltip = 'Save [{}]'
button_exit = 'EXIT'
button_exit_tooltip = 'Exit [{}]'


#########################################################################
# scrapbook.cache
#########################################################################

# map.html
cache_index_toggle_all = 'Toggle all'
cache_index_search_link_title = 'Search'
cache_index_source_link_title = 'Source link'

# search.html
cache_search_title = '%book% :: Search'
cache_search_view_in_map = 'View in map'
cache_search_start = 'go'
cache_search_help_label = 'Search syntax help'
cache_search_help_desc = """\
• Input keywords search from title, content, or comment.
• Use space to separate multiple keywords. For example, “w3c organization” means items containing “w3c” and “organization”.
• Use double quotes to demark a complete phrase, a literal double quote can be escaped by doubling. For example, “"Tom ""loves"" Mary."” means items containing “Tom "loves" Mary.”
• Use minus sign to exclude a keyword. For example, “-virus” means items without “virus”.
• Use “<command>:<keyword>” to specify a special search condition. A minus sign can be prefixed for exclusion or reversion. Available commands include:
  • mc: each subsequent keyword matches case-sensitively. For example, “mc: CIA FBI -mc: president” means items containing “CIA” and “FBI” case-sensitively, and “president” case-insensitively.
  • re: each subsequent keyword is treated as a regular expression. For example, “re: \\bcolou?r\\b -re: 1+1=2” means items that match regular expression “\\bcolou?r\\b” and contain keyword “1+1=2”.
  • id: item whose ID equal to the keyword (or match the regular expression). Multiple values are “or”-connected. For example, “id:2020 id:2021” means items of ID “2020” or “2021”; “-id:2020” means items whose ID is not “2020”.
  • type: items whose type equal to the keyword (or match the regular expression). Multiple values are “or”-connected. Available types are “” (page), “bookmark”, “file”, “note”, etc. For example, “type: type:bookmark” means items whose type is page or bookmark.
  • title: items whose title contains the keyword.
  • content: items whose fulltext contains the keyword.
  • comment: items whose comment contains the keyword.
  • tc: items whose title or comment contains the keyword.
  • tcc: items whose title, fulltext, or comment contains the keyword.
  • index: items whose index file path contains the keyword.
  • source: items whose source URL contains the keyword.
  • icon: items whose icon URL contains the keyword.
  • charset: items whose charset contains the keyword.
  • create: items whose create time matches the condition. Multiple values are “or”-connected. The time condition is an interval with 0-17 digits, followed by a minus sign optionally, and then followed by 0-17 digits. The two 17-digit numbers means the year (4 digits), month (01-12), day (01-31), hours (00-59), minutes (00-59), seconds (00-59), and milliseconds (000-999) in local datetime. Each omitted digit is assumed to be a “0”, except that “999...” is assumed if the end datetime is totally omitted. For example, “create:2014-2015” means since 2014 until 2015; “create:-201309” means before Sep 2013; and “create:20110825” means after Aug 25, 2011.
  • modify: items whose modify time matches the condition. Multiple values are “or”-connected. Time format is same as create.
  • marked: marked items.
  • locked: locked items.
  • location: items with geolocation information.
  • file: items whose filename contains the keyword.
  • root: items under the item of ID. Multiple values are “or”-connected.
  • book: items in the specific scrapbook (by ID). Multiple values are “or”-connected.
  • sort: sort search results using the specific condition, which can be id, title, comment, file, content, source, type, create, or modify. For example, “sort:id -sort:modify” means sorting by ID in acending order and then sorting by modify time in descending order.
  • limit: set a limit on the search result number. For example, “limit:10” means showing the first 10 results, “limit:-20” means removing the last 20 results, and “limit:0” or “-limit:” means unsetting the limit."""
cache_search_result = 'Found %length% results:'
cache_search_result_named = '(%name%) Found %length% results:'
cache_search_sort_last_created = 'Last Created'
cache_search_sort_last_modified = 'Last Modified'
cache_search_sort_title = 'Sort by title'
cache_search_sort_id = 'Sort by ID'
cache_search_confirm_remote = 'Loading remote fulltext cache may require large network flow. Continue?'


#########################################################################
# WebScrapBook
#########################################################################

EditorDeleteAnnotationConfirm = 'Delete this annotation?'



bidi_dir = 'ltr'
html_lang = 'zh'

index_title = '索引於 {path}'

data_table_header_directory = '資料夾'
data_table_header_name = '檔名'
data_table_header_last_modified = '最後修改'
data_table_header_size = '大小'

explorer_tooltip = '檢視模式 [{}]'
explorer_table = '表格檢視'
explorer_gallery = '圖示檢視'
explorer_gallery2 = '圖示檢視 (+多媒體)'

tools_tooltip = '工具 [{}]'
tools_preview_on = '啟用預覽'
tools_preview_off = '停用預覽'
tools_select_all = '全選'
tools_deselect_all = '取消全選'
tools_expand_all = '展開所有項目'
tools_filter = '篩選'
tools_filter_clear = '清除篩選'

command_tooltip = '命令 [{}]'
command_mkdir = '新資料夾'
command_mkzip = '新壓縮檔'
command_mkfile = '新檔案'
command_upload = '上傳'
command_uploaddir = '上傳資料夾'
command_source = '檢視原始碼'
command_download = '下載'
command_exec = '本地執行'
command_browse = '本地檢視'
command_edit = '編輯'
command_editx = '編輯頁面'
command_move = '移動'
command_copy = '複製'
command_link = '建立超連結'
command_delete = '刪除'

command_move_prompt = '移動至此路徑或資料夾:'
command_move_prompt_multi = '移動至此資料夾:'
command_copy_prompt = '複製至此路徑或資料夾:'
command_copy_prompt_multi = '複製至此資料夾:'
command_link_prompt = '建立超連結於此路徑或資料夾:'
command_link_prompt_multi = '建立超連結於此資料夾:'

label_selected = '已選取 %count% 項'

previewer_toolbar_title = """\
鍵盤快捷鍵:
• [Esc]: 關閉預覽。
• [I] 或 [Space]: 切換資訊欄。
• [Enter]: 在新分頁開啟。
• [PageUp] 或 [Alt+←] 或 [←]: 顯示上一項。
• [PageDown] 或 [Alt+→] 或 [→]: 顯示下一項。
• [Home]: 顯示第一項。
• [End]: 顯示最後一項。
• [+] 或 [Alt+↑] 或 [↑]: 放大。
• [-] 或 [Alt+↓] 或 [↓]: 縮小。
• [Ctrl+0]: 縮放至符合可視範圍。
• [Ctrl+1]: 切換縮放為圖片大小與最後縮放大小。
• [Ctrl+←] 或 [Shift+←]: 左移。
• [Ctrl+→] 或 [Shift+→]: 右移。
• [Ctrl+↑] 或 [Shift+↑]: 上移。
• [Ctrl+↓] 或 [Shift+↓]: 下移。
"""
previewer_button_previous = '上一項'
previewer_button_next = '下一項'
previewer_button_infobar = '切換資訊欄'
previewer_button_close = '關閉預覽'

edit_title = '編輯 {path}'
edit_error_no_javascript = '啟用 JavaScript 或升級瀏覽器以支援 JavaScript，此 HTML 編輯器才能正常運作。'
edit_warning_encoding = '此檔案以 {encoding} 解碼，儲存時將編碼為 UTF-8。'

markdown_title = '頁面 {path}'
maff_index_title = '頁面於 {path}'

button_save = '儲存'
button_save_tooltip = '儲存 [{}]'
button_exit = '離開'
button_exit_tooltip = '離開 [{}]'


#########################################################################
# scrapbook.cache
#########################################################################

# map.html
cache_index_toggle_all = '全部展開或收合'
cache_index_search_link_title = '搜尋'
cache_index_source_link_title = '來源連結'

# search.html
cache_search_title = '%book% :: 搜尋'
cache_search_view_in_map = '在索引頁中檢視'
cache_search_start = '查'
cache_search_help_label = '搜尋語法說明'
cache_search_help_desc = """\
• 輸入的關鍵詞會在標題、內文、評註中查找。
• 可用半形空白分隔要檢索的多個關鍵詞。例如「w3c organization」表示搜尋含有「w3c」及「organization」的項目 。
• 可用半形雙引號檢索完整關鍵詞，若關鍵詞包含半形雙引號，可用連續兩個半形雙引號表示。例如「"Tom ""loves"" Mary."」表示搜尋含有「Tom "loves" Mary.」的項目。
• 可用負號排除關鍵詞。例如「-virus」表示搜尋不含「virus」的項目。
• 可用「<命令>:<關鍵詞>」指定特殊的檢索條件，命令前可加負號表示排除或反向。可用命令如下：
  • mc：之後的關鍵詞皆比對大小寫。加負號則反之。例如「mc: CIA FBI -mc: president」表示搜尋含有區分大小寫的「CIA」、「FBI」及不分大小寫的「president」的項目。
  • re：之後的關鍵詞皆視為正規表示式。加負號則反之。例如「re: \\bcolou?r\\b -re: 1+1=2」表示搜尋匹配正規表示式「\\bcolou?r\\b」且含有關鍵詞「1+1=2」的項目。
  • id：搜尋 ID 等同關鍵詞（或與正規表示式匹配）的項目。多次指定時以「或」連接。例如「id:2020 id:2021」表示搜尋 ID 為「2020」或「2021」的項目；「-id:2020」表示搜尋 ID 不為「2020」的項目。
  • type：搜尋類型等同關鍵詞（或與正規表示式匹配）的項目。多次指定時以「或」連接。可用的類型有「」（網頁）、「bookmark」（書籤）、「file」（檔案）、「note」（筆記）等等。例如「type: type:bookmark」表示搜尋類型為網頁或書籤的項目。
  • title：搜尋標題含有關鍵詞的項目。
  • content：搜尋全文含有關鍵詞的項目。
  • comment：搜尋評註含有關鍵詞的項目。
  • tc：搜尋標題或評註含有關鍵詞的項目。
  • tcc：搜尋標題、全文或評註含有關鍵詞的項目。
  • index：搜尋索引檔含有關鍵詞的項目。
  • source：搜尋原始網址含有關鍵詞的項目。
  • icon：搜尋圖示網址含有關鍵詞的項目。
  • charset：搜尋字集含有關鍵詞的項目。
  • create：搜尋建立時間符合條件的項目。多次指定時以「或」連接。時間格式為 0-17 位數字，後面接負號（可略），再接 0-17 位數字，表示時間範圍的起始與結束。兩個 17 位數字表示本地時間的年（4 位數）、月（01-12）、日（01-31）、時（00-59）、分（00-59）、秒（00-59）、毫秒（000-999），省略的部分視為 0，唯結束時間全省略時視為「999...」。例如「create:2014-2015」表示 2014 年到 2015 年，「create:-201309」表示 2013 年九月以前，「create:20110825」表示 2011 年八月 25 日以後。
  • modify：搜尋修改時間符合條件的項目。多次指定時以「或」連接。時間格式同 create。
  • marked：搜尋強調標示的項目。
  • locked：搜尋鎖定的項目。
  • location：搜尋含有地理位置資訊的項目。
  • file：搜尋檔名含有關鍵詞的項目。
  • root：搜尋此 ID 項目下的子項目。多次指定時以「或」連接。
  • book：搜尋指定剪貼簿（依 ID）中的項目。多次指定時以「或」連接。
  • sort：將搜尋結果按指定方式排序，負號表示倒序。可填入 id、title、comment、file、content、source、type、create、modify。例如「sort:id -sort:modify」表示先按 ID 遞增排序再按修改時間遞減排序。
  • limit：設定搜尋結果筆數限制。例如「limit:10」表示只呈現前 10 筆搜尋結果，「limit:-20」表示不呈現最後 20 筆搜尋結果，「limit:0」或「-limit:」表示移除之前設定的限制。"""
cache_search_result = '找到 %length% 筆結果：'
cache_search_result_named = '(%name%) 找到 %length% 筆結果：'
cache_search_sort_last_created = '最後建立'
cache_search_sort_last_modified = '最後修改'
cache_search_sort_title = '標題排序'
cache_search_sort_id = 'ID 排序'
cache_search_confirm_remote = '載入遠端全文快取可能使用較大的網路流量。要繼續嗎？'


#########################################################################
# WebScrapBook
#########################################################################

EditorDeleteAnnotationConfirm = '刪除這個批註嗎？'
